import datetime, os
import pandas as pd
import ibis
from phenex.test.cohort_test_generator import CohortTestGenerator
from phenex.codelists import Codelist
from phenex.phenotypes import (
    AgePhenotype,
    CategoricalPhenotype,
    CodelistPhenotype,
    Cohort,
    TimeRangePhenotype,
    SexPhenotype,
    UserDefinedPhenotype,
)
from phenex.filters import (
    DateFilter,
    ValueFilter,
    CategoricalFilter,
    RelativeTimeRangeFilter,
    GreaterThanOrEqualTo,
    GreaterThan,
    LessThan,
)
from phenex.test.cohort.test_mappings import (
    PersonTableForTests,
    DrugExposureTableForTests,
    ObservationPeriodTableForTests,
    ConditionOccurenceTableForTests,
)

from phenex.test.util.dummy.generate_dummy_cohort_data import generate_dummy_cohort_data


class CohortWithContinuousCoverageTestGenerator(CohortTestGenerator):
    """
    | **entry** | **entry_date** | **obs_start** | **obs_end** | **PATID** | **Comment** |
    | --- | --- | --- | --- | --- | --- |
    | **d1** | 2020-01-01 | 2018-12-30 | 2020-01-10 | P0 | COHORT |
    | **d1** | 2020-01-01 | 2019-01-03 | 2020-01-10 | P2 | obs_start <1 year prior to entry |
    | **d1** | 2020-01-01 | 2018-12-30 | 2019-12-31 | P4 | obs_end before entry_date |
    | **d1** | 2020-01-01 | 2019-01-03 | 2019-12-31 | P6 | obs_start <1 year prior to entry and obs_end before entry_date |
    """

    def define_cohort(self):
        entry = CodelistPhenotype(
            return_date="first",
            codelist=Codelist(["d1"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
        )

        cc = TimeRangePhenotype(
            relative_time_range=RelativeTimeRangeFilter(
                min_days=GreaterThanOrEqualTo(365)
            )
        )

        return Cohort(
            name="test_continuous_coverage",
            entry_criterion=entry,
            inclusions=[cc],
        )

    def generate_dummy_input_data(self):
        values = [
            {
                "name": "entry",
                "values": ["d1", "d4"],
            },
            {
                "name": "entry_date",
                "values": [datetime.date(2020, 1, 1)],
            },
            {
                "name": "obs_start",
                "values": [datetime.date(2018, 12, 30), datetime.date(2019, 1, 3)],
            },
            {
                "name": "obs_end",
                "values": [datetime.date(2020, 1, 10), datetime.date(2019, 12, 31)],
            },
        ]

        return generate_dummy_cohort_data(values)

    def define_mapped_tables(self):
        self.con = ibis.duckdb.connect()
        df_allvalues = self.generate_dummy_input_data()

        # create dummy person table
        df_person = pd.DataFrame(df_allvalues[["PATID"]])
        df_person["YOB"] = 1
        df_person["GENDER"] = 1
        df_person["ACCEPTABLE"] = 1
        schema_person = {"PATID": str, "YOB": int, "GENDER": int, "ACCEPTABLE": int}
        person_table = PersonTableForTests(
            self.con.create_table("PERSON", df_person, schema=schema_person)
        )
        # create drug exposure table
        df_drug_exposure_entry = pd.DataFrame(
            df_allvalues[["PATID", "entry", "entry_date"]]
        )
        # df_drug_exclusion_exposure = pd.DataFrame(df_allvalues[["PATID", "prior_et_use", "prior_et_use_date"]])
        df_drug_exposure_entry.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        # df_drug_exclusion_exposure.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        # df_drug_exposure = pd.concat([df_drug_exposure_entry, df_drug_exclusion_exposure])
        df_drug_exposure = df_drug_exposure_entry

        schema_drug_exposure = {
            "PATID": str,
            "PRODCODEID": str,
            "ISSUEDATE": datetime.date,
        }
        drug_exposure_table = DrugExposureTableForTests(
            self.con.create_table(
                "DRUG_EXPOSURE", df_drug_exposure, schema=schema_drug_exposure
            )
        )

        # create observation period table
        df_obs = pd.DataFrame(df_allvalues[["PATID", "obs_start", "obs_end"]])
        df_obs.columns = ["PATID", "REGSTARTDATE", "REGENDDATE"]

        schema_obs = {
            "PATID": str,
            "REGSTARTDATE": datetime.date,
            "REGENDDATE": datetime.date,
        }
        obs_table = ObservationPeriodTableForTests(
            self.con.create_table("OBSERVATION_PERIOD", df_obs, schema=schema_obs)
        )
        return {
            "PERSON": person_table,
            "DRUG_EXPOSURE": drug_exposure_table,
            "OBSERVATION_PERIOD": obs_table,
        }

    def define_expected_output(self):
        df_expected_index = pd.DataFrame()
        df_expected_index["PERSON_ID"] = ["P0"]
        test_infos = {
            "index": df_expected_index,
        }
        return test_infos


class CohortWithContinuousCoverageAndExclusionTestGenerator(CohortTestGenerator):
    """
    | **PATID** | **entry** | **entry_date** | **obs_start** | **obs_end** | **prior_et_use** | **prior_et_use_date** |
    | --- | --- | --- | --- | --- | --- | --- |
    | **P0** | d1 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e1 | 2019-04-01 |
    | **P1** | d4 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e1 | 2019-04-01 |
    | **P2** | d1 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e1 | 2019-04-01 |
    | **P3** | d4 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e1 | 2019-04-01 |
    | **P4** | d1 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e1 | 2019-04-01 |
    | **P5** | d4 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e1 | 2019-04-01 |
    | **P6** | d1 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e1 | 2019-04-01 |
    | **P7** | d4 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e1 | 2019-04-01 |
    | **P8** | d1 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e4 | 2019-04-01 |
    | **P9** | d4 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e4 | 2019-04-01 |
    | **P10** | d1 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e4 | 2019-04-01 |
    | **P11** | d4 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e4 | 2019-04-01 |
    | **P12** | d1 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e4 | 2019-04-01 |
    | **P13** | d4 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e4 | 2019-04-01 |
    | **P14** | d1 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e4 | 2019-04-01 |
    | **P15** | d4 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e4 | 2019-04-01 |
    """

    def define_cohort(self):
        entry = CodelistPhenotype(
            return_date="first",
            codelist=Codelist(["d1"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
        )

        cc = TimeRangePhenotype(
            relative_time_range=RelativeTimeRangeFilter(
                min_days=GreaterThanOrEqualTo(365)
            )
        )

        e4 = CodelistPhenotype(
            name="prior_et_usage",
            codelist=Codelist(["e4"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
            relative_time_range=RelativeTimeRangeFilter(
                when="before", min_days=GreaterThanOrEqualTo(0)
            ),
        )

        def user_defined_function(mapped_tables):
            table = mapped_tables["DRUG_EXPOSURE"]
            table = table.filter(table.PRODCODEID == "d1")
            table = table.select("PERSON_ID")
            return table

        udp = UserDefinedPhenotype(name="udp", function=user_defined_function)

        return Cohort(
            name="test_continuous_coverage_with_exclusion",
            entry_criterion=entry,
            inclusions=[cc, udp],
            exclusions=[e4],
        )

    def generate_dummy_input_data(self):
        values = [
            {
                "name": "entry",
                "values": ["d1", "d4"],
            },
            {
                "name": "entry_date",
                "values": [datetime.date(2020, 1, 1)],
            },
            {
                "name": "obs_start",
                "values": [datetime.date(2018, 12, 30), datetime.date(2019, 1, 3)],
            },
            {
                "name": "obs_end",
                "values": [datetime.date(2020, 1, 10), datetime.date(2019, 12, 31)],
            },
            {"name": "prior_et_use", "values": ["e1", "e4"]},
            {
                "name": "prior_et_use_date",
                "values": [datetime.date(2019, 4, 1)],
            },
        ]

        return generate_dummy_cohort_data(values)

    def define_mapped_tables(self):
        self.con = ibis.duckdb.connect()
        df_allvalues = self.generate_dummy_input_data()

        # create dummy person table
        df_person = pd.DataFrame(df_allvalues[["PATID"]])
        df_person["YOB"] = 1
        df_person["GENDER"] = 1
        df_person["ACCEPTABLE"] = 1
        schema_person = {"PATID": str, "YOB": int, "GENDER": int, "ACCEPTABLE": int}
        person_table = PersonTableForTests(
            self.con.create_table("PERSON", df_person, schema=schema_person)
        )
        # create drug exposure table
        df_drug_exposure_entry = pd.DataFrame(
            df_allvalues[["PATID", "entry", "entry_date"]]
        )
        df_drug_exclusion_exposure = pd.DataFrame(
            df_allvalues[["PATID", "prior_et_use", "prior_et_use_date"]]
        )
        df_drug_exposure_entry.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        df_drug_exclusion_exposure.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        df_drug_exposure = pd.concat(
            [df_drug_exposure_entry, df_drug_exclusion_exposure]
        )

        schema_drug_exposure = {
            "PATID": str,
            "PRODCODEID": str,
            "ISSUEDATE": datetime.date,
        }
        drug_exposure_table = DrugExposureTableForTests(
            self.con.create_table(
                "DRUG_EXPOSURE", df_drug_exposure, schema=schema_drug_exposure
            )
        )

        # create observation period table
        df_obs = pd.DataFrame(df_allvalues[["PATID", "obs_start", "obs_end"]])
        df_obs.columns = ["PATID", "REGSTARTDATE", "REGENDDATE"]

        schema_obs = {
            "PATID": str,
            "REGSTARTDATE": datetime.date,
            "REGENDDATE": datetime.date,
        }
        obs_table = ObservationPeriodTableForTests(
            self.con.create_table("OBSERVATION_PERIOD", df_obs, schema=schema_obs)
        )
        return {
            "PERSON": person_table,
            "DRUG_EXPOSURE": drug_exposure_table,
            "OBSERVATION_PERIOD": obs_table,
        }

    def define_expected_output(self):
        df_expected_index = pd.DataFrame()
        df_expected_index["PERSON_ID"] = ["P0"]
        test_infos = {
            "index": df_expected_index,
        }
        return test_infos


class CohortWithContinuousCoverageExclusionAndAgeTestGenerator(CohortTestGenerator):
    """
    | **PATID** | **entry** | **entry_date** | **obs_start** | **obs_end** | **prior_et_use** | **prior_et_use_date** | **YOB** |
    | --- | --- | --- | --- | --- | --- | --- | --- |
    | **P0** | d1 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e1 | 2019-04-01 | 2000 |
    | **P1** | d4 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e1 | 2019-04-01 | 2000 |
    | **P2** | d1 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e1 | 2019-04-01 | 2000 |
    | **P3** | d4 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e1 | 2019-04-01 | 2000 |
    | **P4** | d1 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e1 | 2019-04-01 | 2000 |
    | **P5** | d4 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e1 | 2019-04-01 | 2000 |
    | **P6** | d1 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e1 | 2019-04-01 | 2000 |
    | **P7** | d4 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e1 | 2019-04-01 | 2000 |
    | **P8** | d1 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e4 | 2019-04-01 | 2000 |
    | **P9** | d4 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e4 | 2019-04-01 | 2000 |
    | **P10** | d1 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e4 | 2019-04-01 | 2000 |
    | **P11** | d4 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e4 | 2019-04-01 | 2000 |
    | **P12** | d1 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e4 | 2019-04-01 | 2000 |
    | **P13** | d4 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e4 | 2019-04-01 | 2000 |
    | **P14** | d1 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e4 | 2019-04-01 | 2000 |
    | **P15** | d4 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e4 | 2019-04-01 | 2000 |
    | **P16** | d1 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e1 | 2019-04-01 | 2010 |
    | **P17** | d4 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e1 | 2019-04-01 | 2010 |
    | **P18** | d1 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e1 | 2019-04-01 | 2010 |
    | **P19** | d4 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e1 | 2019-04-01 | 2010 |
    | **P20** | d1 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e1 | 2019-04-01 | 2010 |
    | **P21** | d4 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e1 | 2019-04-01 | 2010 |
    | **P22** | d1 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e1 | 2019-04-01 | 2010 |
    | **P23** | d4 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e1 | 2019-04-01 | 2010 |
    | **P24** | d1 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e4 | 2019-04-01 | 2010 |
    | **P25** | d4 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e4 | 2019-04-01 | 2010 |
    | **P26** | d1 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e4 | 2019-04-01 | 2010 |
    | **P27** | d4 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e4 | 2019-04-01 | 2010 |
    | **P28** | d1 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e4 | 2019-04-01 | 2010 |
    | **P29** | d4 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e4 | 2019-04-01 | 2010 |
    | **P30** | d1 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e4 | 2019-04-01 | 2010 |
    | **P31** | d4 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e4 | 2019-04-01 | 2010 |
    """

    def define_cohort(self):
        entry = CodelistPhenotype(
            return_date="first",
            codelist=Codelist(["d1"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
        )

        cc = TimeRangePhenotype(
            relative_time_range=RelativeTimeRangeFilter(
                min_days=GreaterThanOrEqualTo(365)
            )
        )
        agege18 = AgePhenotype(
            value_filter=ValueFilter(min_value=GreaterThanOrEqualTo(18))
        )

        e4 = CodelistPhenotype(
            name="prior_et_usage",
            codelist=Codelist(["e4"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
            relative_time_range=RelativeTimeRangeFilter(
                when="before", min_days=GreaterThanOrEqualTo(0)
            ),
        )

        return Cohort(
            name="test_continuous_coverage_age_with_exclusion",
            entry_criterion=entry,
            inclusions=[cc, agege18],
            exclusions=[e4],
        )

    def generate_dummy_input_data(self):
        values = [
            {
                "name": "entry",
                "values": ["d1", "d4"],
            },
            {
                "name": "entry_date",
                "values": [datetime.date(2020, 1, 1)],
            },
            {
                "name": "obs_start",
                "values": [datetime.date(2018, 12, 30), datetime.date(2019, 1, 3)],
            },
            {
                "name": "obs_end",
                "values": [datetime.date(2020, 1, 10), datetime.date(2019, 12, 31)],
            },
            {"name": "prior_et_use", "values": ["e1", "e4"]},
            {
                "name": "prior_et_use_date",
                "values": [datetime.date(2019, 4, 1)],
            },
            {"name": "YOB", "values": [2000, 2010]},
        ]

        return generate_dummy_cohort_data(values)

    def define_mapped_tables(self):
        self.con = ibis.duckdb.connect()
        df_allvalues = self.generate_dummy_input_data()

        # create dummy person table
        df_person = pd.DataFrame(df_allvalues[["PATID", "YOB"]])
        df_person["GENDER"] = 1
        df_person["ACCEPTABLE"] = 1
        schema_person = {"PATID": str, "YOB": int, "GENDER": int, "ACCEPTABLE": int}
        person_table = PersonTableForTests(
            self.con.create_table("PERSON", df_person, schema=schema_person)
        )
        # create drug exposure table
        df_drug_exposure_entry = pd.DataFrame(
            df_allvalues[["PATID", "entry", "entry_date"]]
        )
        df_drug_exclusion_exposure = pd.DataFrame(
            df_allvalues[["PATID", "prior_et_use", "prior_et_use_date"]]
        )
        df_drug_exposure_entry.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        df_drug_exclusion_exposure.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        df_drug_exposure = pd.concat(
            [df_drug_exposure_entry, df_drug_exclusion_exposure]
        )

        schema_drug_exposure = {
            "PATID": str,
            "PRODCODEID": str,
            "ISSUEDATE": datetime.date,
        }
        drug_exposure_table = DrugExposureTableForTests(
            self.con.create_table(
                "DRUG_EXPOSURE", df_drug_exposure, schema=schema_drug_exposure
            )
        )
        # create observation period table
        df_obs = pd.DataFrame(df_allvalues[["PATID", "obs_start", "obs_end"]])
        df_obs.columns = ["PATID", "REGSTARTDATE", "REGENDDATE"]

        schema_obs = {
            "PATID": str,
            "REGSTARTDATE": datetime.date,
            "REGENDDATE": datetime.date,
        }
        obs_table = ObservationPeriodTableForTests(
            self.con.create_table("OBSERVATION_PERIOD", df_obs, schema=schema_obs)
        )
        return {
            "PERSON": person_table,
            "DRUG_EXPOSURE": drug_exposure_table,
            "OBSERVATION_PERIOD": obs_table,
        }

    def define_expected_output(self):
        df_expected_index = pd.DataFrame()
        df_expected_index["PERSON_ID"] = ["P0"]
        test_infos = {
            "index": df_expected_index,
        }
        return test_infos


class CohortWithContinuousCoverageExclusionAndAgeAsExclusionTestGenerator(
    CohortWithContinuousCoverageExclusionAndAgeTestGenerator
):
    """
    Identical output and cohort definition to CohortWithContinuousCoverageExclusionAndAgeTestGenerator however using inverse of age. Instead of inclusion age >= 18, use exclusion age<18
    """

    def define_cohort(self):
        entry = CodelistPhenotype(
            return_date="first",
            codelist=Codelist(["d1"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
        )

        cc = TimeRangePhenotype(
            relative_time_range=RelativeTimeRangeFilter(
                min_days=GreaterThanOrEqualTo(365)
            )
        )
        agel18 = AgePhenotype(value_filter=ValueFilter(max_value=LessThan(18)))

        e4 = CodelistPhenotype(
            name="prior_et_usage",
            codelist=Codelist(["e4"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
            relative_time_range=RelativeTimeRangeFilter(
                when="before", min_days=GreaterThanOrEqualTo(0)
            ),
        )

        return Cohort(
            name="test_continuous_coverage_age_with_exclusion",
            entry_criterion=entry,
            inclusions=[cc],
            exclusions=[e4, agel18],
        )

    def generate_dummy_input_data(self):
        values = [
            {
                "name": "entry",
                "values": ["d1", "d4"],
            },
            {
                "name": "entry_date",
                "values": [datetime.date(2020, 1, 1)],
            },
            {
                "name": "obs_start",
                "values": [datetime.date(2018, 12, 30), datetime.date(2019, 1, 3)],
            },
            {
                "name": "obs_end",
                "values": [datetime.date(2020, 1, 10), datetime.date(2019, 12, 31)],
            },
            {"name": "prior_et_use", "values": ["e1", "e4"]},
            {
                "name": "prior_et_use_date",
                "values": [datetime.date(2019, 4, 1)],
            },
            {"name": "YOB", "values": [2000, 2010]},
        ]

        return generate_dummy_cohort_data(values)


class CohortWithContinuousCoverageExclusionAgeSexTestGenerator(CohortTestGenerator):
    """
    | **PATID** | **entry** | **entry_date** | **obs_start** | **obs_end** | **prior_et_use** | **prior_et_use_date** | **YOB** | **GENDER** |
    | --- | --- | --- | --- | --- | --- | --- | --- | --- |
    | **P0** | d1 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e1 | 2019-04-01 | 2000 | 1 |
    | **P1** | d4 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e1 | 2019-04-01 | 2000 | 1 |
    | **P2** | d1 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e1 | 2019-04-01 | 2000 | 1 |
    | **P3** | d4 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e1 | 2019-04-01 | 2000 | 1 |
    | **P4** | d1 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e1 | 2019-04-01 | 2000 | 1 |
    | **P5** | d4 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e1 | 2019-04-01 | 2000 | 1 |
    | **P6** | d1 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e1 | 2019-04-01 | 2000 | 1 |
    | **P7** | d4 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e1 | 2019-04-01 | 2000 | 1 |
    | **P8** | d1 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e4 | 2019-04-01 | 2000 | 1 |
    | **P9** | d4 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e4 | 2019-04-01 | 2000 | 1 |
    | **P10** | d1 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e4 | 2019-04-01 | 2000 | 1 |
    | **P11** | d4 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e4 | 2019-04-01 | 2000 | 1 |
    | **P12** | d1 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e4 | 2019-04-01 | 2000 | 1 |
    | **P13** | d4 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e4 | 2019-04-01 | 2000 | 1 |
    | **P14** | d1 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e4 | 2019-04-01 | 2000 | 1 |
    | **P15** | d4 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e4 | 2019-04-01 | 2000 | 1 |
    | **P16** | d1 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e1 | 2019-04-01 | 2010 | 1 |
    | **P17** | d4 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e1 | 2019-04-01 | 2010 | 1 |
    | **P18** | d1 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e1 | 2019-04-01 | 2010 | 1 |
    | **P19** | d4 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e1 | 2019-04-01 | 2010 | 1 |
    | **P20** | d1 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e1 | 2019-04-01 | 2010 | 1 |
    | **P21** | d4 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e1 | 2019-04-01 | 2010 | 1 |
    | **P22** | d1 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e1 | 2019-04-01 | 2010 | 1 |
    | **P23** | d4 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e1 | 2019-04-01 | 2010 | 1 |
    | **P24** | d1 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e4 | 2019-04-01 | 2010 | 1 |
    | **P25** | d4 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e4 | 2019-04-01 | 2010 | 1 |
    | **P26** | d1 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e4 | 2019-04-01 | 2010 | 1 |
    | **P27** | d4 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e4 | 2019-04-01 | 2010 | 1 |
    | **P28** | d1 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e4 | 2019-04-01 | 2010 | 1 |
    | **P29** | d4 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e4 | 2019-04-01 | 2010 | 1 |
    | **P30** | d1 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e4 | 2019-04-01 | 2010 | 1 |
    | **P31** | d4 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e4 | 2019-04-01 | 2010 | 1 |
    | **P32** | d1 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e1 | 2019-04-01 | 2000 | 2 |
    | **P33** | d4 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e1 | 2019-04-01 | 2000 | 2 |
    | **P34** | d1 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e1 | 2019-04-01 | 2000 | 2 |
    | **P35** | d4 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e1 | 2019-04-01 | 2000 | 2 |
    | **P36** | d1 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e1 | 2019-04-01 | 2000 | 2 |
    | **P37** | d4 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e1 | 2019-04-01 | 2000 | 2 |
    | **P38** | d1 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e1 | 2019-04-01 | 2000 | 2 |
    | **P39** | d4 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e1 | 2019-04-01 | 2000 | 2 |
    | **P40** | d1 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e4 | 2019-04-01 | 2000 | 2 |
    | **P41** | d4 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e4 | 2019-04-01 | 2000 | 2 |
    | **P42** | d1 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e4 | 2019-04-01 | 2000 | 2 |
    | **P43** | d4 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e4 | 2019-04-01 | 2000 | 2 |
    | **P44** | d1 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e4 | 2019-04-01 | 2000 | 2 |
    | **P45** | d4 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e4 | 2019-04-01 | 2000 | 2 |
    | **P46** | d1 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e4 | 2019-04-01 | 2000 | 2 |
    | **P47** | d4 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e4 | 2019-04-01 | 2000 | 2 |
    | **P48** | d1 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e1 | 2019-04-01 | 2010 | 2 |
    | **P49** | d4 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e1 | 2019-04-01 | 2010 | 2 |
    | **P50** | d1 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e1 | 2019-04-01 | 2010 | 2 |
    | **P51** | d4 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e1 | 2019-04-01 | 2010 | 2 |
    | **P52** | d1 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e1 | 2019-04-01 | 2010 | 2 |
    | **P53** | d4 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e1 | 2019-04-01 | 2010 | 2 |
    | **P54** | d1 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e1 | 2019-04-01 | 2010 | 2 |
    | **P55** | d4 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e1 | 2019-04-01 | 2010 | 2 |
    | **P56** | d1 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e4 | 2019-04-01 | 2010 | 2 |
    | **P57** | d4 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e4 | 2019-04-01 | 2010 | 2 |
    | **P58** | d1 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e4 | 2019-04-01 | 2010 | 2 |
    | **P59** | d4 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e4 | 2019-04-01 | 2010 | 2 |
    | **P60** | d1 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e4 | 2019-04-01 | 2010 | 2 |
    | **P61** | d4 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e4 | 2019-04-01 | 2010 | 2 |
    | **P62** | d1 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e4 | 2019-04-01 | 2010 | 2 |
    | **P63** | d4 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e4 | 2019-04-01 | 2010 | 2 |
    """

    def define_cohort(self):
        entry = CodelistPhenotype(
            return_date="first",
            codelist=Codelist(["d1"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
        )

        cc = TimeRangePhenotype(
            relative_time_range=RelativeTimeRangeFilter(
                min_days=GreaterThanOrEqualTo(365)
            )
        )
        agege18 = AgePhenotype(
            value_filter=ValueFilter(min_value=GreaterThanOrEqualTo(18))
        )
        sex = SexPhenotype(categorical_filter=CategoricalFilter(allowed_values=[1]))

        e4 = CodelistPhenotype(
            name="prior_et_usage",
            codelist=Codelist(["e4"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
            relative_time_range=RelativeTimeRangeFilter(
                when="before", min_days=GreaterThanOrEqualTo(0)
            ),
        )

        return Cohort(
            name="test_continuous_coverage_age_sex_with_exclusion",
            entry_criterion=entry,
            inclusions=[cc, agege18, sex],
            exclusions=[e4],
        )

    def generate_dummy_input_data(self):
        values = [
            {
                "name": "entry",
                "values": ["d1", "d4"],
            },
            {
                "name": "entry_date",
                "values": [datetime.date(2020, 1, 1)],
            },
            {
                "name": "obs_start",
                "values": [datetime.date(2018, 12, 30), datetime.date(2019, 1, 3)],
            },
            {
                "name": "obs_end",
                "values": [datetime.date(2020, 1, 10), datetime.date(2019, 12, 31)],
            },
            {"name": "prior_et_use", "values": ["e1", "e4"]},
            {
                "name": "prior_et_use_date",
                "values": [datetime.date(2019, 4, 1)],
            },
            {"name": "YOB", "values": [2000, 2010]},
            {"name": "GENDER", "values": [1, 2]},
        ]

        return generate_dummy_cohort_data(values)

    def define_mapped_tables(self):
        self.con = ibis.duckdb.connect()
        df_allvalues = self.generate_dummy_input_data()

        # create dummy person table
        df_person = pd.DataFrame(df_allvalues[["PATID", "YOB", "GENDER"]])
        df_person["ACCEPTABLE"] = 1
        schema_person = {"PATID": str, "YOB": int, "GENDER": int, "ACCEPTABLE": int}
        person_table = PersonTableForTests(
            self.con.create_table("PERSON", df_person, schema=schema_person)
        )
        # create drug exposure table
        df_drug_exposure_entry = pd.DataFrame(
            df_allvalues[["PATID", "entry", "entry_date"]]
        )
        df_drug_exclusion_exposure = pd.DataFrame(
            df_allvalues[["PATID", "prior_et_use", "prior_et_use_date"]]
        )
        df_drug_exposure_entry.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        df_drug_exclusion_exposure.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        df_drug_exposure = pd.concat(
            [df_drug_exposure_entry, df_drug_exclusion_exposure]
        )

        schema_drug_exposure = {
            "PATID": str,
            "PRODCODEID": str,
            "ISSUEDATE": datetime.date,
        }
        drug_exposure_table = DrugExposureTableForTests(
            self.con.create_table(
                "DRUG_EXPOSURE", df_drug_exposure, schema=schema_drug_exposure
            )
        )
        # create observation period table
        df_obs = pd.DataFrame(df_allvalues[["PATID", "obs_start", "obs_end"]])
        df_obs.columns = ["PATID", "REGSTARTDATE", "REGENDDATE"]

        schema_obs = {
            "PATID": str,
            "REGSTARTDATE": datetime.date,
            "REGENDDATE": datetime.date,
        }
        obs_table = ObservationPeriodTableForTests(
            self.con.create_table("OBSERVATION_PERIOD", df_obs, schema=schema_obs)
        )
        return {
            "PERSON": person_table,
            "DRUG_EXPOSURE": drug_exposure_table,
            "OBSERVATION_PERIOD": obs_table,
        }

    def define_expected_output(self):
        df_expected_index = pd.DataFrame()
        df_expected_index["PERSON_ID"] = ["P0"]
        test_infos = {
            "index": df_expected_index,
        }
        return test_infos


class CohortWithUDPTestGenerator(CohortTestGenerator):
    """
    | **PATID** | **entry** | **entry_date** | **obs_start** | **obs_end** | **prior_et_use** | **prior_et_use_date** |
    | --- | --- | --- | --- | --- | --- | --- |
    | **P0** | d1 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e1 | 2019-04-01 |
    | **P1** | d4 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e1 | 2019-04-01 |
    | **P2** | d1 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e1 | 2019-04-01 |
    | **P3** | d4 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e1 | 2019-04-01 |
    | **P4** | d1 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e1 | 2019-04-01 |
    | **P5** | d4 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e1 | 2019-04-01 |
    | **P6** | d1 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e1 | 2019-04-01 |
    | **P7** | d4 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e1 | 2019-04-01 |
    | **P8** | d1 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e4 | 2019-04-01 |
    | **P9** | d4 | 2020-01-01 | 2018-12-30 | 2020-01-10 | e4 | 2019-04-01 |
    | **P10** | d1 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e4 | 2019-04-01 |
    | **P11** | d4 | 2020-01-01 | 2019-01-03 | 2020-01-10 | e4 | 2019-04-01 |
    | **P12** | d1 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e4 | 2019-04-01 |
    | **P13** | d4 | 2020-01-01 | 2018-12-30 | 2019-12-31 | e4 | 2019-04-01 |
    | **P14** | d1 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e4 | 2019-04-01 |
    | **P15** | d4 | 2020-01-01 | 2019-01-03 | 2019-12-31 | e4 | 2019-04-01 |
    """

    def define_cohort(self):

        def user_defined_function(mapped_tables):
            df = pd.DataFrame()
            df["PERSON_ID"] = ["P0", "P1", "P2", "P3", "P4"]
            df["EVENT_DATE"] = [
                datetime.date(2020, 1, 1),  # d1 give correct cc
                datetime.date(2020, 1, 1),
                datetime.date(2020, 1, 3),  # d1 give correct cc
                datetime.date(2020, 1, 1),
                datetime.date(2018, 1, 1),  # d1 give incorrect cc
            ]
            return ibis.memtable(df)

        entry = UserDefinedPhenotype(name="udp", function=user_defined_function)

        d1 = CodelistPhenotype(
            return_date="first",
            codelist=Codelist(["d1"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
        )

        cc = TimeRangePhenotype(
            relative_time_range=RelativeTimeRangeFilter(
                min_days=GreaterThanOrEqualTo(30)
            )
        )

        return Cohort(
            name="test_udp_entry",
            entry_criterion=entry,
            inclusions=[cc, d1],
        )

    def generate_dummy_input_data(self):
        values = [
            {
                "name": "entry",
                "values": ["d1", "d4"],
            },
            {
                "name": "entry_date",
                "values": [datetime.date(2020, 1, 1)],
            },
            {
                "name": "obs_start",
                "values": [datetime.date(2018, 12, 30), datetime.date(2019, 1, 3)],
            },
            {
                "name": "obs_end",
                "values": [datetime.date(2020, 1, 10), datetime.date(2019, 12, 31)],
            },
            {"name": "prior_et_use", "values": ["e1", "e4"]},
            {
                "name": "prior_et_use_date",
                "values": [datetime.date(2019, 4, 1)],
            },
        ]

        return generate_dummy_cohort_data(values)

    def define_mapped_tables(self):
        self.con = ibis.duckdb.connect()
        df_allvalues = self.generate_dummy_input_data()

        # create dummy person table
        df_person = pd.DataFrame(df_allvalues[["PATID"]])
        df_person["YOB"] = 1
        df_person["GENDER"] = 1
        df_person["ACCEPTABLE"] = 1
        schema_person = {"PATID": str, "YOB": int, "GENDER": int, "ACCEPTABLE": int}
        person_table = PersonTableForTests(
            self.con.create_table("PERSON", df_person, schema=schema_person)
        )
        # create drug exposure table
        df_drug_exposure_entry = pd.DataFrame(
            df_allvalues[["PATID", "entry", "entry_date"]]
        )
        df_drug_exclusion_exposure = pd.DataFrame(
            df_allvalues[["PATID", "prior_et_use", "prior_et_use_date"]]
        )
        df_drug_exposure_entry.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        df_drug_exclusion_exposure.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        df_drug_exposure = pd.concat(
            [df_drug_exposure_entry, df_drug_exclusion_exposure]
        )

        schema_drug_exposure = {
            "PATID": str,
            "PRODCODEID": str,
            "ISSUEDATE": datetime.date,
        }
        drug_exposure_table = DrugExposureTableForTests(
            self.con.create_table(
                "DRUG_EXPOSURE", df_drug_exposure, schema=schema_drug_exposure
            )
        )

        # create observation period table
        df_obs = pd.DataFrame(df_allvalues[["PATID", "obs_start", "obs_end"]])
        df_obs.columns = ["PATID", "REGSTARTDATE", "REGENDDATE"]

        schema_obs = {
            "PATID": str,
            "REGSTARTDATE": datetime.date,
            "REGENDDATE": datetime.date,
        }
        obs_table = ObservationPeriodTableForTests(
            self.con.create_table("OBSERVATION_PERIOD", df_obs, schema=schema_obs)
        )
        return {
            "PERSON": person_table,
            "DRUG_EXPOSURE": drug_exposure_table,
            "OBSERVATION_PERIOD": obs_table,
        }

    def define_expected_output(self):
        df_expected_index = pd.DataFrame()
        df_expected_index["PERSON_ID"] = ["P0", "P2"]
        test_infos = {
            "index": df_expected_index,
        }
        return test_infos


def test_time_range_phenotype():
    g = CohortWithContinuousCoverageTestGenerator()
    g.run_tests()


def test_continuous_coverage_with_exclusion():
    g = CohortWithContinuousCoverageAndExclusionTestGenerator()
    g.run_tests()


def test_continuous_coverage_age_with_exclusion():
    g = CohortWithContinuousCoverageExclusionAndAgeTestGenerator()
    g.run_tests()


def test_continuous_coverage_age_as_exclusion_with_exclusion():
    g = CohortWithContinuousCoverageExclusionAndAgeAsExclusionTestGenerator()
    g.run_tests()


def test_continuous_coverage_age_sex_with_exclusion():
    g = CohortWithContinuousCoverageExclusionAgeSexTestGenerator()
    g.run_tests()


def test_udp_entry():
    g = CohortWithUDPTestGenerator()
    g.run_tests()


if __name__ == "__main__":
    test_udp_entry()



from typing import List
from datetime import date
from ibis.expr.types.relations import Table
import ibis
from phenex.tables import PhenexTable


def attach_anchor_and_get_reference_date(table, anchor_phenotype=None):
    if anchor_phenotype is not None:
        if anchor_phenotype.table is None:
            raise ValueError(
                f"Dependent Phenotype {anchor_phenotype.name} must be executed before this node can run!"
            )
        else:
            anchor_table = anchor_phenotype.table
            reference_column = anchor_table.EVENT_DATE
            # Note that joins can change column names if the tables have name collisions!
            table = table.join(anchor_table, "PERSON_ID")
    else:
        assert (
            "INDEX_DATE" in table.columns
        ), f"INDEX_DATE column not found in table {table}"
        reference_column = table.INDEX_DATE

    return table, reference_column


def hstack(phenotypes: List["Phenotype"], join_table: Table = None) -> Table:
    """
    Horizontally stacks multiple PhenotypeTable objects into a single table. The PERSON_ID columns are used to join the tables together. The resulting table will have three columns per phenotype: BOOLEAN, EVENT_DATE, and VALUE. The columns will be contain the phenotype name as a prefix.
    # TODO: Add a test for this function.
    Args:
        phenotypes (List[Phenotype]): A list of Phenotype objects to stack.
    """
    # TODO decide if phenotypes should be returning a phenextable
    if isinstance(join_table, PhenexTable):
        join_table = join_table.table
    idx_phenotype_to_begin = 0
    join_type = "left"  # if join table is defined, we want to left join
    if join_table is None:
        idx_phenotype_to_begin = 1
        join_table = phenotypes[0].namespaced_table
        join_type = "outer"  # if join table is NOT defined, we want an outer join
    for pt in phenotypes[idx_phenotype_to_begin:]:
        join_table = join_table.join(pt.namespaced_table, "PERSON_ID", how=join_type)
        join_table = join_table.mutate(
            PERSON_ID=ibis.coalesce(join_table.PERSON_ID, join_table.PERSON_ID_right)
        )
        columns = join_table.columns
        columns.remove("PERSON_ID_right")
        join_table = join_table.select(columns)

    for pt in phenotypes:
        column_operation = join_table[f"{pt.name}_BOOLEAN"].fill_null(False)
        join_table = join_table.mutate(**{f"{pt.name}_BOOLEAN": column_operation})
    return join_table


def select_phenotype_columns(
    table, fill_date=ibis.null(date), fill_value=ibis.null(), fill_boolean=True
):
    if "PERSON_ID" not in table.columns:
        raise ValueError("Table must have a PERSON_ID column")
    if "EVENT_DATE" not in table.columns:
        table = table.mutate(EVENT_DATE=fill_date)
    if "VALUE" not in table.columns:
        table = table.mutate(VALUE=fill_value)
    if "BOOLEAN" not in table.columns:
        table = table.mutate(BOOLEAN=fill_boolean)
    return table.select([table.PERSON_ID, table.BOOLEAN, table.EVENT_DATE, table.VALUE])



import yaml
import os
import pandas as pd
import ibis

from phenex.reporting import InExCounts, Waterfall
from .util.check_equality import check_counts_table_equal, check_equality
from phenex.util import create_logger

logger = create_logger(__name__)


def sort_by_personid(_df):
    _df["id"] = [int(x.replace("P", "")) for x in _df["PERSON_ID"].values]
    return _df.sort_values(by="id").drop("id", axis=1)


class CohortTestGenerator:
    """
    This class is a base class for all TestGenerators.

    FIXME Document how to subclass and use.
    """

    date_format = "%m-%d-%Y"
    test_values = False
    test_date = False

    def __init__(self):
        pass

    def run_tests(self, path="phenex/test/cohort", verbose=False):
        self.verbose = verbose
        self.cohort = self.define_cohort()

        self._create_artifact_directory(self.cohort.name, path)
        self.mapped_tables = self.define_mapped_tables()
        self._write_mapped_tables()
        self._generate_output_artifacts()
        self._run_tests()

    def define_cohort(self):
        raise NotImplementedError

    def define_mapped_tables(self):
        raise NotImplementedError

    def define_expected_output(self):
        raise NotImplementedError

    def name_file(self, test_info):
        return f"{self.cohort.name}__{test_info['name']}"

    def name_output_file(self, test_info):
        return self.name_file(test_info) + "_output"

    def _generate_output_artifacts(self):
        self.test_infos = self.define_expected_output()
        for test_name, df in self.test_infos.items():
            filename = test_name + ".csv"
            path = os.path.join(self.dirpaths["expected"], filename)
            df.to_csv(path, index=False, date_format=self.date_format)

    def _write_mapped_tables(self):
        for domain, table in self.mapped_tables.items():
            path = os.path.join(self.dirpaths["mapped_tables"], f"{domain}.csv")
            table.to_pandas().to_csv(path, index=False)

    def _run_tests(self):
        self.cohort.execute(self.mapped_tables)

        logger.debug(
            f"\nENTRY\n{sort_by_personid(self.cohort.entry_criterion.table.to_pandas())}"
        )
        if len(self.cohort.inclusions) > 0:
            logger.debug(
                f"\nINCLUSIONS\n{sort_by_personid(self.cohort.inclusions_table.to_pandas())}"
            )

        if len(self.cohort.exclusions) > 0:
            logger.debug(
                f"\nEXCLUSIONS\n{sort_by_personid(self.cohort.exclusions_table.to_pandas())}"
            )
        r = Waterfall()
        logger.debug(f"\nWATERFALL\n{r.execute(self.cohort)}")
        logger.debug(
            f"\nINDEX\n{sort_by_personid(self.cohort.index_table.to_pandas())}"
        )

        if (
            "counts_inclusion" in self.test_infos.keys()
            or "counts_exclusion" in self.test_infos.keys()
        ):
            self._test_inex_counts()

        if "index" in self.test_infos.keys():
            self._test_index_table(
                result=self.cohort.index_table, expected=self.test_infos["index"]
            )

    def _create_artifact_directory(self, name_demo, path):
        if os.path.exists(path):
            path_artifacts = os.path.join(path, "artifacts")
        else:
            raise ValueError(
                "Pass a path to the cohort test generator where expected and calculated output should be written"
            )
        path_cohort = os.path.join(path_artifacts, name_demo)

        self.dirpaths = {
            "artifacts": path_artifacts,
            "cohort": path_cohort,
            "expected": os.path.join(path_cohort, "expected"),
            "result": os.path.join(path_cohort, "result"),
            "mapped_tables": os.path.join(path_cohort, "mapped_tables"),
        }
        for _path in self.dirpaths.values():
            if not os.path.exists(_path):
                os.makedirs(_path)

    def _test_index_table(self, result, expected):
        name = self.cohort.name + "_index"
        expected_output_table = self.con.create_table(name, expected)
        join_on = ["PERSON_ID"]
        if self.test_values:
            join_on.append("VALUE")
        if self.test_date:
            join_on.append("EVENT_DATE")
        check_equality(
            result,
            expected_output_table,
            test_name=self.cohort.name + "_index",
            test_values=self.test_values,
            test_date=self.test_date,
            join_on=join_on,
        )

    def _test_inex_counts(self):
        # compute inclusion exclusion counts
        r = InExCounts()
        r.execute(self.cohort)
        # write expected results to artifacts
        r.df_counts_inclusion.to_csv(
            os.path.join(self.dirpaths["result"], "counts_inclusion.csv"), index=False
        )
        r.df_counts_exclusion.to_csv(
            os.path.join(self.dirpaths["result"], "counts_exclusion.csv"), index=False
        )
        # test results
        if len(self.cohort.inclusions) > 0:
            check_counts_table_equal(
                result=r.df_counts_inclusion,
                expected=self.test_infos["counts_inclusion"],
                test_name=self.cohort.name + "_inclusion",
            )
        if len(self.cohort.exclusions) > 0:
            check_counts_table_equal(
                result=r.df_counts_exclusion,
                expected=self.test_infos["counts_exclusion"],
                test_name=self.cohort.name + "_exclusion",
            )



from typing import List, Dict, Optional
from phenex.phenotypes.phenotype import Phenotype
import ibis
from ibis.expr.types.relations import Table
from phenex.tables import PhenotypeTable
from phenex.phenotypes.functions import hstack
from phenex.reporting import Table1
from phenex.util import create_logger
from concurrent.futures import ThreadPoolExecutor

logger = create_logger(__name__)


def subset_and_add_index_date(tables: Dict[str, Table], index_table: PhenotypeTable):
    subset_tables = {}
    for key, table in tables.items():
        columns = list(set(["INDEX_DATE"] + table.columns))
        subset_tables[key] = type(table)(
            table.inner_join(index_table, "PERSON_ID").select(columns)
        )
    return subset_tables


class Cohort(Phenotype):
    """
    The Cohort class represents a cohort of individuals based on specified entry criteria,
    inclusions, exclusions, and baseline characteristics. It extends the Phenotype class.

    Parameters:
        name: A descriptive name for the cohort.
        entry_criterion: The phenotype used to define index date for the cohort.
        inclusions: A list of phenotypes that must evaluate to True for patients to be included in the cohort.
        exclusions: A list of phenotypes that must evaluate to False for patients to be included in the cohort.
        characteristics: A list of phenotypes representing baseline characteristics of the cohort to be computed for all patients passing the inclusion and exclusion criteria.
        outcomes: A list of phenotypes representing outcomes of the cohort.

    Attributes:
        table (PhenotypeTable): The resulting index table after filtering (None until execute is called)
        inclusions_table (Table): The patient-level result of all inclusion criteria calculations (None until execute is called)
        exclusions_table (Table): The patient-level result of all exclusion criteria calculations (None until execute is called)
        characteristics_table (Table): The patient-level result of all baseline characteristics caclulations. (None until execute is called)
        outcomes_table (Table): The patient-level result of all outcomes caclulations. (None until execute is called)
        subset_tables_entry (Dict[str, PhenexTable]): Tables that have been subset by those patients satisfying the entry criterion.
        subset_tables_index (Dict[str, PhenexTable]): Tables that have been subset by those patients satisfying the entry, inclusion and exclusion criteria.
    """

    table = None

    def __init__(
        self,
        entry_criterion: Phenotype,
        inclusions: Optional[List[Phenotype]] = None,
        exclusions: Optional[List[Phenotype]] = None,
        characteristics: Optional[List[Phenotype]] = None,
        derived_tables: Optional[List["DerivedTable"]] = None,
        outcomes: Optional[List[Phenotype]] = None,
        **kwargs,
    ):
        super(Cohort, self).__init__(**kwargs)
        self.entry_criterion = entry_criterion
        self.inclusions = inclusions if inclusions is not None else []
        self.exclusions = exclusions if exclusions is not None else []
        self.characteristics = characteristics if characteristics is not None else []
        self.outcomes = outcomes if outcomes is not None else []
        self.index_table = None
        self.exclusions_table = None
        self.inclusions_table = None
        self.characteristics_table = None
        self.outcomes_table = None
        self.derived_tables = derived_tables
        self.children = (
            [entry_criterion]
            + self.inclusions
            + self.exclusions
            + self.characteristics
            + self.outcomes
        )
        self._table1 = None
        logger.info(
            f"Cohort '{self.name}' initialized with entry criterion '{self.entry_criterion.name}'"
        )

    def execute(
        self,
        tables: Dict[str, Table],
        con: "SnowflakeConnector" = None,
        write_subset_tables=False,
        overwrite: bool = False,
        n_threads: int = 1,
    ) -> PhenotypeTable:
        """
        The execute method executes the full cohort in order of computation. The order is entry criterion -> inclusion -> exclusion -> baseline characteristics. Tables are subset at two points, after entry criterion and after full inclusion/exclusion calculation to result in subset_entry data (contains all source data for patients that fulfill the entry criterion, with a possible index date) and subset_index data (contains all source data for patients that fulfill all in/ex criteria, with a set index date). Additionally, default reporters are executed such as table 1 for baseline characteristics.

        Args:
            tables (Dict[str, Table]): A dictionary of table names to Table objects.
            con (SnowflakeConnector, optional): A connection to Snowflake. Defaults to None. If passed, will write entry, inclusions, exclusions, index, characteristics and outcomes tables.
            write_subset_tables (bool, optional): Whether to write subset tables (subset-entry and subset-index) in addition to the standard intermediate tables.
            overwrite (bool, optional): Whether to overwrite existing tables when writing to disk.
            n_threads (int, optional): Number of threads to use for parallel execution. Defaults to 1.

        Returns:
            PhenotypeTable: The index table corresponding the cohort.
        """
        logger.info(f"Executing cohort '{self.name}' with {n_threads} threads...")
        # Compute entry criterion
        logger.debug("Computing entry criterion ...")
        self.entry_criterion.execute(tables)
        if con:
            logger.debug("Writing entry table ...")
            self.entry_criterion.table = con.create_table(
                self.entry_criterion.table, f"{self.name}__entry", overwrite=overwrite
            )

        logger.debug("Entry criterion computed.")
        index_table = self.entry_criterion.table.mutate(INDEX_DATE="EVENT_DATE")

        self.subset_tables_entry = subset_and_add_index_date(tables, index_table)
        if write_subset_tables:
            logger.debug("Writing subset entry tables ...")
            with ThreadPoolExecutor(max_workers=n_threads) as executor:
                futures = {}
                for key, table in self.subset_tables_entry.items():
                    futures[key] = executor.submit(
                        con.create_table,
                        table.table,
                        f"{self.name}__subset_entry_{key}".upper(),
                        overwrite,
                    )
                for key, future in futures.items():
                    self.subset_tables_entry[key] = type(self.subset_tables_entry[key])(
                        future.result()
                    )

        self.subset_tables_entry = self.derive_tables(
            self.subset_tables_entry, index_table, con, overwrite=overwrite
        )

        # Apply inclusions if any
        if self.inclusions:
            logger.debug("Applying inclusions ...")
            self._compute_inclusions_table(n_threads)
            if con:
                logger.debug("Writing inclusions table ...")
                self.inclusions_table = con.create_table(
                    self.inclusions_table,
                    f"{self.name}__inclusions".upper(),
                    overwrite=overwrite,
                )
            include = self.inclusions_table.filter(
                self.inclusions_table["BOOLEAN"] == True
            ).select(["PERSON_ID"])
            index_table = index_table.inner_join(include, ["PERSON_ID"])
            logger.debug("Inclusions applied.")

        # Apply exclusions if any
        if self.exclusions:
            logger.debug("Applying exclusions ...")
            self._compute_exclusions_table(n_threads)
            if con:
                logger.debug("Writing exclusions table ...")
                self.exclusions_table = con.create_table(
                    self.exclusions_table,
                    f"{self.name}__exclusions".upper(),
                    overwrite=overwrite,
                )
            exclude = self.exclusions_table.filter(
                self.exclusions_table["BOOLEAN"] == False
            ).select(["PERSON_ID"])
            index_table = index_table.inner_join(exclude, ["PERSON_ID"])
            logger.debug("Exclusions applied.")

        self.index_table = index_table
        if con:
            logger.debug("Writing index table ...")
            self.index_table = con.create_table(
                index_table, f"{self.name}__index".upper(), overwrite=overwrite
            )

        self.subset_tables_index = subset_and_add_index_date(
            self.subset_tables_entry, self.index_table.select("PERSON_ID")
        )
        if write_subset_tables:
            logger.debug("Writing subset index tables ...")
            with ThreadPoolExecutor(max_workers=n_threads) as executor:
                futures = {}
                for key, table in self.subset_tables_index.items():
                    futures[key] = executor.submit(
                        con.create_table,
                        table.table,
                        f"{self.name}__subset_index_{key}".upper(),
                        overwrite,
                    )
                for key, future in futures.items():
                    self.subset_tables_index[key] = type(self.subset_tables_index[key])(
                        future.result()
                    )
            logger.debug("Finished writing subset index tables ...")

        if self.characteristics:
            logger.debug("Computing characteristics ...")
            self._compute_characteristics_table(n_threads)
            if con:
                logger.debug("Writing characteristics table ...")
                self.characteristics_table = con.create_table(
                    self.characteristics_table,
                    f"{self.name}__characteristics".upper(),
                    overwrite=overwrite,
                )
            logger.debug("Characteristics computed.")
            _ = self.table1

        if self.outcomes:
            logger.debug("Computing outcomes ...")
            self._compute_outcomes_table(n_threads)
            logger.debug("Outcomes computed.")

        logger.info(f"Cohort '{self.name}' execution completed.")
        return index_table

    def _compute_inclusions_table(self, n_threads: int) -> Table:
        logger.debug("Computing inclusions table")
        """
        Compute the inclusions table from the individual inclusions phenotypes.
        Meant only to be called internally from execute() so that all dependent phenotypes
        have already been computed.

        Returns:
            Table: The join of all inclusion phenotypes together with a single "BOOLEAN"
            column that is the logical AND of all individual inclusion phenotypes
        """
        # create an inex table;
        # rows are persons that fulfill the entry criterion
        # columns are inclusion criteria with true of false if that column pt criteria are fulfilled
        inclusions_table = self._compute_inex_table(self.inclusions, n_threads)

        # create the final boolean inclusion column
        # this is true only if all inclusion criteria are true
        inclusions_table = inclusions_table.mutate(
            BOOLEAN=ibis.least(
                *[inclusions_table[f"{x.name}_BOOLEAN"] for x in self.inclusions]
            )
        )
        self.inclusions_table = inclusions_table
        logger.debug("Inclusions table computed")
        return self.inclusions_table

    def _compute_exclusions_table(self, n_threads: int) -> Table:
        logger.debug("Computing exclusions table")
        """
        Compute the exclusions table from the individual exclusions phenotypes.

        Returns:
            Table: The join of all exclusions phenotypes together with a single "BOOLEAN"
            column that is the logical OR of all individual inclusion phenotypes
        """
        # create an inex table;
        # rows are persons that fulfill the entry criterion
        # columns are inclusion criteria with true of false if fulfill
        exclusions_table = self._compute_inex_table(self.exclusions, n_threads)

        # create the boolean inclusions column
        # this is true only if all inclusions criteria are true
        exclusions_table = exclusions_table.mutate(
            BOOLEAN=ibis.greatest(
                *[exclusions_table[f"{x.name}_BOOLEAN"] for x in self.exclusions]
            )
        )
        self.exclusions_table = exclusions_table
        logger.debug("Exclusions table computed")
        return self.exclusions_table

    def _compute_inex_table(
        self, phenotypes: List["Phenotype"], n_threads: int
    ) -> Table:
        logger.debug("Computing inex table")
        """
        Compute the exclusion table from the individual exclusion phenotypes.

        Returns:
            Table: The join of all inclusion phenotypes together with a single "BOOLEAN"
            column that is the logical AND of all individual inclusion phenotypes
        """
        inex_table = self.entry_criterion.table.select(["PERSON_ID"])
        # execute all phenotypes and join the boolean column only
        with ThreadPoolExecutor(max_workers=n_threads) as executor:
            futures = [
                executor.submit(pt.execute, self.subset_tables_entry)
                for pt in phenotypes
            ]
            for future in futures:
                future.result()
        for pt in phenotypes:
            pt_table = pt.table.select(["PERSON_ID", "BOOLEAN"]).rename(
                **{
                    f"{pt.name}_BOOLEAN": "BOOLEAN",
                }
            )
            inex_table = inex_table.left_join(pt_table, ["PERSON_ID"])
            columns = inex_table.columns
            columns.remove("PERSON_ID_right")
            inex_table = inex_table.select(columns)

        # fill all nones with False
        boolean_columns = [col for col in inex_table.columns if "BOOLEAN" in col]
        for col in boolean_columns:
            inex_table = inex_table.mutate({col: inex_table[col].fill_null(False)})
        logger.debug("Inex table computed")
        return inex_table

    def _compute_characteristics_table(self, n_threads: int) -> Table:
        logger.debug("Computing characteristics table")
        """
        Retrieves and joins all characteristic tables.
        Meant only to be called internally from _execute() so that all dependent phenotypes
        have already been computed.

        Returns:
            Table: The join of all characteristic tables.
        """
        with ThreadPoolExecutor(max_workers=n_threads) as executor:
            futures = [
                executor.submit(c.execute, self.subset_tables_index)
                for c in self.characteristics
            ]
            for future in futures:
                future.result()
        self.characteristics_table = hstack(
            self.characteristics,
            join_table=self.index_table.select(["PERSON_ID", "INDEX_DATE"]),
        )
        logger.debug("Characteristics table computed")
        return self.characteristics_table

    def _compute_outcomes_table(self, n_threads: int) -> Table:
        logger.debug("Computing outcomes table")
        """
        Retrieves and joins all outcome tables. Meant only to be called internally from execute() so that all dependent phenotypes have already been computed.

        Returns:
            Table: The join of all outcome tables.
        """
        with ThreadPoolExecutor(max_workers=n_threads) as executor:
            futures = [
                executor.submit(o.execute, self.subset_tables_index)
                for o in self.outcomes
            ]
            for future in futures:
                future.result()
        self.outcomes_table = hstack(
            self.outcomes,
            join_table=self.index_table.select(["PERSON_ID", "INDEX_DATE"]),
        )
        logger.debug("Outcomes table computed")
        return self.outcomes_table

    @property
    def table1(self):
        if self._table1 is None:
            logger.debug("Generating Table1 report ...")
            reporter = Table1()
            self._table1 = reporter.execute(self)
            logger.debug("Table1 report generated.")
        return self._table1

    def derive_tables(self, tables, index_table, con, overwrite=False):
        if self.derived_tables is None:
            return tables
        for derived_table in self.derived_tables:
            logger.info(f"Deriving table {derived_table.dest_domain}")
            # derive the table
            table = derived_table.execute(
                tables=tables,
            )
            # join on index table to get index date
            columns = ["INDEX_DATE"] + table.columns
            table = type(table)(
                table.inner_join(index_table, "PERSON_ID").select(columns)
            )

            if con:
                logger.info(
                    f"Saving derived table {self.name}__subset_entry_{derived_table.dest_domain}"
                )
                table_name = f"{self.name}__subset_entry_{derived_table.dest_domain}"
                tables[derived_table.dest_domain] = type(table)(
                    con.create_table(table, table_name, overwrite=overwrite)
                )
            else:
                tables[derived_table.dest_domain] = table
        return tables



from .phenotype import Phenotype

from .codelist_phenotype import CodelistPhenotype
from .age_phenotype import AgePhenotype
from .bin_phenotype import BinPhenotype
from .sex_phenotype import SexPhenotype
from .event_count_phenotype import EventCountPhenotype
from .measurement_phenotype import MeasurementPhenotype
from .measurement_change_phenotype import MeasurementChangePhenotype
from .death_phenotype import DeathPhenotype
from .categorical_phenotype import CategoricalPhenotype
from .time_range_phenotype import TimeRangePhenotype
from .user_defined_phenotype import UserDefinedPhenotype
from .computation_graph_phenotypes import (
    ScorePhenotype,
    ArithmeticPhenotype,
    LogicPhenotype,
)
from .within_same_encounter_phenotype import WithinSameEncounterPhenotype
from .cohort import Cohort

# from .age_phenotype import AgePhenotype
# from .event_count_phenotype import EventCountPhenotype

# # from .arithmetic_phenotype import ArithmeticPhenotype



import pandas as pd

from phenex.reporting.reporter import Reporter
from phenex.util import create_logger
from ibis import _

logger = create_logger(__name__)


class Table1(Reporter):
    """
    Table1 is a common term used in epidemiology to describe a table that shows an overview of the baseline characteristics of a cohort. It contains the counts and percentages of the cohort that have each characteristic, for both boolean and value characteristics. In addition, summary statistics are provided for value characteristics (mean, std, median, min, max).

    """

    def execute(self, cohort: "Cohort") -> pd.DataFrame:
        if len(cohort.characteristics) == 0:
            logger.info("No characteristics. table1 is empty")
            return pd.DataFrame()

        self.cohort = cohort
        self.cohort_names_in_order = [x.name for x in self.cohort.characteristics]
        self.N = (
            cohort.index_table.filter(cohort.index_table.BOOLEAN == True)
            .select("PERSON_ID")
            .distinct()
            .count()
            .execute()
        )
        logger.debug("Starting with categorical columns for table1")
        self.df_categoricals = self._report_categorical_columns()
        logger.debug("Starting with boolean columns for table1")
        self.df_booleans = self._report_boolean_columns()
        logger.debug("Starting with value columns for table1")
        self.df_values = self._report_value_columns()

        # add the full cohort size as the first row
        df_n = pd.DataFrame({"N": [self.N], "inex_order": [-1]}, index=["Cohort"])
        # add percentage column
        dfs = [
            df
            for df in [df_n, self.df_booleans, self.df_values, self.df_categoricals]
            if df is not None
        ]
        if len(dfs) > 1:
            self.df = pd.concat(dfs)
        elif len(dfs) == 1:
            self.df = dfs[0]
        else:
            self.df = None
        if self.df is not None:
            self.df["%"] = 100 * self.df["N"] / self.N
            # reorder columns so N and % are first
            first_cols = ["N", "%"]
            column_order = first_cols + [
                x for x in self.df.columns if x not in first_cols
            ]
            self.df = self.df[column_order]
        logger.debug("Finished creating table1")

        if self.pretty_display:
            self.create_pretty_display()

        self.df = self.df.sort_values(by=["inex_order", "Name"])
        self.df = self.df.reset_index()[
            [x for x in self.df.columns if x not in ["index", "inex_order"]]
        ]
        return self.df

    def _get_boolean_characteristics(self):
        return [
            x
            for x in self.cohort.characteristics
            if type(x).__name__
            not in [
                "MeasurementPhenotype",
                "AgePhenotype",
                "TimeRangePhenotype",
                "ScorePhenotype",
                "CategoricalPhenotype",
                "SexPhenotype",
                "ArithmeticPhenotype",
                "EventCountPhenotype",
                "BinPhenotype",
            ]
        ]

    def _get_value_characteristics(self):
        return [
            x
            for x in self.cohort.characteristics
            if type(x).__name__
            in [
                "MeasurementPhenotype",
                "AgePhenotype",
                "TimeRangePhenotype",
                "ArithmeticPhenotype",
                "EventCountPhenotype",  # event count is a value; show summary statistics for number of days
            ]
        ]

    def _get_categorical_characteristics(self):
        return [
            x
            for x in self.cohort.characteristics
            if type(x).__name__
            in [
                "CategoricalPhenotype",
                "SexPhenotype",
                "ScorePhenotype",  # score is categorical; show number of patients in each score category
                "BinPhenotype",
            ]
        ]

    def _get_boolean_count_for_phenotype(self, phenotype):
        return (
            phenotype.table.select(["PERSON_ID", "BOOLEAN"])
            .distinct()["BOOLEAN"]
            .sum()
            .execute()
        )

    def _report_boolean_columns(self):
        table = self.cohort.characteristics_table
        # get list of all boolean columns
        boolean_phenotypes = self._get_boolean_characteristics()
        logger.debug(
            f"Found {len(boolean_phenotypes)} : {[x.name for x in boolean_phenotypes]}"
        )
        if len(boolean_phenotypes) == 0:
            return None
        # get count of 'Trues' in the boolean columns i.e. the phenotype counts
        df_t1 = pd.DataFrame()
        df_t1["N"] = [
            self._get_boolean_count_for_phenotype(phenotype)
            for phenotype in boolean_phenotypes
        ]
        df_t1.index = [
            x.display_name if self.pretty_display else x.name
            for x in boolean_phenotypes
        ]
        df_t1["inex_order"] = [
            self.cohort_names_in_order.index(x.name) for x in boolean_phenotypes
        ]
        return df_t1

    def _report_value_columns(self):
        value_phenotypes = self._get_value_characteristics()
        logger.debug(
            f"Found {len(value_phenotypes)} : {[x.name for x in value_phenotypes]}"
        )

        if len(value_phenotypes) == 0:
            return None

        names = []
        dfs = []
        for phenotype in value_phenotypes:
            _table = phenotype.table.select(["PERSON_ID", "VALUE"]).distinct()
            d = {
                "N": self._get_boolean_count_for_phenotype(phenotype),
                "Mean": _table["VALUE"].mean().execute(),
                "STD": _table["VALUE"].std().execute(),
                "Median": _table["VALUE"].median().execute(),
                "Min": _table["VALUE"].min().execute(),
                "Max": _table["VALUE"].max().execute(),
                "inex_order": self.cohort_names_in_order.index(phenotype.name),
            }
            dfs.append(pd.DataFrame.from_dict([d]))
            names.append(
                phenotype.display_name if self.pretty_display else phenotype.name
            )
        if len(dfs) == 1:
            df = dfs[0]
        else:
            df = pd.concat(dfs)
        df.index = names
        return df

    def _report_categorical_columns(self):
        categorical_phenotypes = self._get_categorical_characteristics()
        logger.debug(
            f"Found {len(categorical_phenotypes)} : {[x.name for x in categorical_phenotypes]}"
        )
        if len(categorical_phenotypes) == 0:
            return None
        dfs = []
        names = []
        for phenotype in categorical_phenotypes:
            name = phenotype.display_name if self.pretty_display else phenotype.name
            _table = phenotype.table.select(["PERSON_ID", "VALUE"])
            # Get counts for each category
            cat_counts = (
                _table.distinct().group_by("VALUE").aggregate(N=_.count()).execute()
            )
            cat_counts.index = [f"{name}={v}" for v in cat_counts["VALUE"]]
            _df = pd.DataFrame(cat_counts["N"])
            _df["inex_order"] = self.cohort_names_in_order.index(phenotype.name)
            dfs.append(_df)
            names.extend(cat_counts.index)
        if len(dfs) == 1:
            df = dfs[0]
        else:
            df = pd.concat(dfs)
        df.index = names
        return df

    def create_pretty_display(self):
        # cast counts to integer and to str, so that we can display without 'NaNs'
        self.df["N"] = self.df["N"].astype("Int64").astype(str)
        self.df = self.df.reset_index()
        self.df.columns = ["Name"] + list(self.df.columns[1:])

        self.df = self.df.round(self.decimal_places)

        to_prettify = ["%", "Mean", "STD", "Median", "Min", "Max"]
        for column in to_prettify:
            self.df[column] = self.df[column].astype(str)

        self.df = self.df.replace("<NA>", "").replace("nan", "")



from typing import Dict
from ibis.expr.types.relations import Table
from phenex.phenotypes.phenotype import Phenotype
from phenex.tables import PhenotypeTable, is_phenex_phenotype_table


class MockPhenotype(Phenotype):
    """
    Used to mock the execution of a Phenotype for testing purposes.
    """

    def __init__(self, table):
        self.children = []
        assert is_phenex_phenotype_table(table)
        super(MockPhenotype, self).__init__()
        self.table = table

    def _execute(self, tables: Dict[str, Table]):
        """
        Executes the phenotype processing logic.

        Args:
            tables (Dict[str, Table]): A dictionary where the keys are table names and the values are Table objects.

        Raises:
            NotImplementedError: This method should be implemented by subclasses.
        """
        return self.table



from typing import Optional, Union
from phenex.filters.filter import Filter
from phenex.tables import PhenexTable
from phenex.filters.value import (
    Value,
    GreaterThan,
    GreaterThanOrEqualTo,
    LessThan,
    LessThanOrEqualTo,
)


class ValueFilter(Filter):
    """
    ValueFilter filters events in an PhenexTable based on a specified value range.

    Parameters:
        min_value: Minimum value required to pass through the filter.
        max_value: Maximum value required to pass through the filter.
        column_name: The column name to which the value range should be applied. Default to VALUE, which is the default name of the value column in PhenotypeTable's.

    Methods:
        filter: Filters the given PhenexTable based on the range of values specified by the min_value and max_value attributes. See Filter.
    """

    def __init__(
        self,
        min_value: Optional[Union[Value, GreaterThan, GreaterThanOrEqualTo]] = None,
        max_value: Optional[Union[Value, LessThan, LessThanOrEqualTo]] = None,
        column_name: Optional[str] = "VALUE",
    ):
        if min_value is not None:
            assert min_value.operator in [
                ">",
                ">=",
            ], f"min_value operator must be > or >=, not {min_value.operator}"
        if max_value is not None:
            assert max_value.operator in [
                "<",
                "<=",
            ], f"max_value operator must be > or >=, not {max_value.operator}"
        if max_value is not None and min_value is not None:
            assert (
                min_value.value <= max_value.value
            ), f"min_value must be less than or equal to max_value"
        self.min_value = min_value
        self.max_value = max_value
        self.column_name = column_name
        super(ValueFilter, self).__init__()

    def _filter(self, table: PhenexTable) -> PhenexTable:
        conditions = []
        value_column = getattr(table, self.column_name)
        if self.min_value is not None:
            if self.min_value.operator == ">":
                conditions.append(value_column > self.min_value.value)
            elif self.min_value.operator == ">=":
                conditions.append(value_column >= self.min_value.value)
            else:
                raise ValueError("Operator for min_value days be > or >=")
        if self.max_value is not None:
            if self.max_value.operator == "<":
                conditions.append(value_column < self.max_value.value)
            elif self.max_value.operator == "<=":
                conditions.append(value_column <= self.max_value.value)
            else:
                raise ValueError("Operator for max_value days be < or <=")
        if conditions:
            table = table.filter(conditions)
        return table



import datetime, os
import pandas as pd
import ibis
from phenex.test.cohort_test_generator import CohortTestGenerator
from phenex.codelists import Codelist
from phenex.phenotypes import (
    AgePhenotype,
    CategoricalPhenotype,
    CodelistPhenotype,
    Cohort,
    TimeRangePhenotype,
    SexPhenotype,
)
from phenex.filters import *
from phenex.test.cohort.test_mappings import (
    PersonTableForTests,
    DrugExposureTableForTests,
    ObservationPeriodTableForTests,
    ConditionOccurenceTableForTests,
)

from phenex.test.util.dummy.generate_dummy_cohort_data import generate_dummy_cohort_data


class SimpleCohortWithExclusionTestGenerator(CohortTestGenerator):
    def define_cohort(self):
        entry = CodelistPhenotype(
            return_date="first",
            codelist=Codelist(["d1"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
        )

        e4 = CodelistPhenotype(
            name="prior_et_usage",
            codelist=Codelist(["e4"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
            relative_time_range=RelativeTimeRangeFilter(
                when="before", min_days=GreaterThanOrEqualTo(0)
            ),
        )

        return Cohort(
            name="test_simple_cohort_with_exclusion",
            entry_criterion=entry,
            exclusions=[e4],
        )

    def generate_dummy_input_data(self):
        values = [
            {
                "name": "entry",
                "values": ["d1", "d4"],
            },
            {
                "name": "entry_date",
                "values": [datetime.date(2020, 1, 1)],
            },
            {"name": "prior_et_use", "values": ["e1", "e4"]},
            {
                "name": "prior_et_use_date",
                "values": [datetime.date(2019, 4, 1)],
            },
        ]

        return generate_dummy_cohort_data(values)

    def define_mapped_tables(self):
        self.con = ibis.duckdb.connect()
        df_allvalues = self.generate_dummy_input_data()

        # create dummy person table
        df_person = pd.DataFrame(df_allvalues[["PATID"]])
        df_person["YOB"] = 1
        df_person["GENDER"] = 1
        df_person["ACCEPTABLE"] = 1
        schema_person = {"PATID": str, "YOB": int, "GENDER": int, "ACCEPTABLE": int}
        person_table = PersonTableForTests(
            self.con.create_table("PERSON", df_person, schema=schema_person)
        )
        # create drug exposure table
        df_drug_exposure_entry = pd.DataFrame(
            df_allvalues[["PATID", "entry", "entry_date"]]
        )
        df_drug_exclusion_exposure = pd.DataFrame(
            df_allvalues[["PATID", "prior_et_use", "prior_et_use_date"]]
        )
        df_drug_exposure_entry.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        df_drug_exclusion_exposure.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        df_drug_exposure = pd.concat(
            [df_drug_exposure_entry, df_drug_exclusion_exposure]
        )

        schema_drug_exposure = {
            "PATID": str,
            "PRODCODEID": str,
            "ISSUEDATE": datetime.date,
        }
        drug_exposure_table = DrugExposureTableForTests(
            self.con.create_table(
                "DRUG_EXPOSURE", df_drug_exposure, schema=schema_drug_exposure
            )
        )
        return {
            "PERSON": person_table,
            "DRUG_EXPOSURE": drug_exposure_table,
        }

    def define_expected_output(self):
        df_expected_index = pd.DataFrame()
        df_expected_index["PERSON_ID"] = ["P0"]
        test_infos = {
            "index": df_expected_index,
        }
        return test_infos


class SimpleCohortWithExclusionAndStudyPeriodTestGenerator(
    SimpleCohortWithExclusionTestGenerator
):
    """
    Adding a study period to our study; there are now patients with an entry event prior to the study period. Check that they are properly removed
    """

    def define_cohort(self):
        study_period = DateFilter(
            min_date=AfterOrOn(datetime.date(2015, 1, 1)),
            max_date=BeforeOrOn(datetime.date(2020, 12, 31)),
        )

        entry = CodelistPhenotype(
            return_date="first",
            codelist=Codelist(["d1"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
            date_range=study_period,
        )

        e4 = CodelistPhenotype(
            name="prior_et_usage",
            codelist=Codelist(["e4"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
            relative_time_range=RelativeTimeRangeFilter(
                when="before", min_days=GreaterThanOrEqualTo(0)
            ),
        )

        return Cohort(
            name="test_simple_cohort_with_exclusion_and_study_period",
            entry_criterion=entry,
            exclusions=[e4],
        )

    def generate_dummy_input_data(self):
        values = [
            {
                "name": "entry",
                "values": ["d1", "d4"],
            },
            {
                "name": "entry_date",
                "values": [datetime.date(2020, 1, 1), datetime.date(2010, 1, 1)],
            },
            {"name": "prior_et_use", "values": ["e1", "e4"]},
            {
                "name": "prior_et_use_date",
                "values": [datetime.date(2019, 4, 1)],
            },
        ]
        return generate_dummy_cohort_data(values)


class SimpleCohortWithExclusionPostIndexTestGenerator(
    SimpleCohortWithExclusionTestGenerator
):
    """
    Same cohort as test_study_period. Here add new test data with exclusion events occurring AFTER index date (these patients should NOT be excluded)

    | **entry** | **entry_date** | **prior_et_use** | **prior_et_use_date** | **PATID** |  |
    | --- | --- | --- | --- | --- | --- |
    | **d1** | 2020-01-01 | e1 | 2019-04-01 | P0 | COHORT : no exclusion event (e4) |
    | **d1** | 2010-01-01 | e1 | 2019-04-01 | P2 | Entry date prior to study period |
    | **d1** | 2020-01-01 | e4 | 2019-04-01 | P4 | Fulfills exclusion |
    | **d1** | 2010-01-01 | e4 | 2019-04-01 | P6 | Entry date prior to study period |
    | **d1** | 2020-01-01 | e1 | 2021-04-01 | P8 | COHORT : no exclusion event (e4) |
    | **d1** | 2010-01-01 | e1 | 2021-04-01 | P10 | Entry date prior to study period |
    | **d1** | 2020-01-01 | e4 | 2021-04-01 | P12 | COHORT : exclusion event after index |
    | **d1** | 2010-01-01 | e4 | 2021-04-01 | P14 | Entry date prior to study period |
    """

    def define_cohort(self):
        study_period = DateFilter(
            min_date=AfterOrOn(datetime.date(2015, 1, 1)),
            max_date=BeforeOrOn(datetime.date(2020, 12, 31)),
        )

        entry = CodelistPhenotype(
            return_date="first",
            codelist=Codelist(["d1"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
            date_range=study_period,
        )

        e4 = CodelistPhenotype(
            name="prior_et_usage",
            codelist=Codelist(["e4"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
            relative_time_range=RelativeTimeRangeFilter(
                when="before", min_days=GreaterThanOrEqualTo(0)
            ),
        )
        return Cohort(
            name="test_simple_cohort_with_exclusions_post_index",
            entry_criterion=entry,
            exclusions=[e4],
        )

    def generate_dummy_input_data(self):
        values = [
            {
                "name": "entry",
                "values": ["d1", "d4"],
            },
            {
                "name": "entry_date",
                "values": [datetime.date(2020, 1, 1), datetime.date(2010, 1, 1)],
            },
            {"name": "prior_et_use", "values": ["e1", "e4"]},
            {
                "name": "prior_et_use_date",
                "values": [datetime.date(2019, 4, 1), datetime.date(2021, 4, 1)],
            },
        ]
        return generate_dummy_cohort_data(values)

    def define_expected_output(self):
        df_expected_index = pd.DataFrame()
        df_expected_index["PERSON_ID"] = ["P0", "P8", "P12"]
        test_infos = {
            "index": df_expected_index,
        }
        return test_infos


class SimpleCohortWithInclusionTestGenerator(SimpleCohortWithExclusionTestGenerator):
    def define_cohort(self):
        entry = CodelistPhenotype(
            return_date="first",
            codelist=Codelist(["d1"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
        )

        i1 = CodelistPhenotype(
            name="prior_et_usage",
            codelist=Codelist(["e1"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
            relative_time_range=RelativeTimeRangeFilter(
                when="before", min_days=GreaterThanOrEqualTo(0)
            ),
        )
        return Cohort(
            name="test_simple_cohort_with_inclusions",
            entry_criterion=entry,
            inclusions=[i1],
        )


class SimpleCohortWithInclusionAndExclusionTestGenerator(CohortTestGenerator):
    def define_cohort(self):
        entry = CodelistPhenotype(
            return_date="first",
            codelist=Codelist(["d1"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
        )

        i1 = CodelistPhenotype(
            name="prior_et_usage",
            codelist=Codelist(["i1"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
            relative_time_range=RelativeTimeRangeFilter(
                when="before", min_days=GreaterThanOrEqualTo(0)
            ),
        )

        e4 = CodelistPhenotype(
            name="prior_et_usage",
            codelist=Codelist(["e4"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
            relative_time_range=RelativeTimeRangeFilter(
                when="before", min_days=GreaterThanOrEqualTo(0)
            ),
        )

        return Cohort(
            name="test_simple_cohort_with_inclusions",
            entry_criterion=entry,
            inclusions=[i1],
            exclusions=[e4],
        )

    def generate_dummy_input_data(self):
        values = [
            {
                "name": "entry",
                "values": ["d1", "d4"],
            },
            {
                "name": "entry_date",
                "values": [datetime.date(2020, 1, 1), datetime.date(2010, 1, 1)],
            },
            {"name": "exclusions", "values": ["e1", "e4"]},
            {
                "name": "exclusions_date",
                "values": [datetime.date(2019, 4, 1)],
            },
            {"name": "inclusions", "values": ["i1", "i4"]},
            {
                "name": "inclusions_date",
                "values": [datetime.date(2019, 4, 1)],
            },
        ]
        return generate_dummy_cohort_data(values)

    def define_expected_output(self):
        df_expected_index = pd.DataFrame()
        df_expected_index["PERSON_ID"] = ["P0"]
        test_infos = {
            "index": df_expected_index,
        }
        return test_infos

    def define_mapped_tables(self):
        self.con = ibis.duckdb.connect()
        df_allvalues = self.generate_dummy_input_data()

        # create dummy person table
        df_person = pd.DataFrame(df_allvalues[["PATID"]])
        df_person["YOB"] = 1
        df_person["GENDER"] = 1
        df_person["ACCEPTABLE"] = 1
        schema_person = {"PATID": str, "YOB": int, "GENDER": int, "ACCEPTABLE": int}
        person_table = PersonTableForTests(
            self.con.create_table("PERSON", df_person, schema=schema_person)
        )
        # create drug exposure table
        df_drug_exposure_entry = pd.DataFrame(
            df_allvalues[["PATID", "entry", "entry_date"]]
        )
        df_drug_inclusion_exposure = pd.DataFrame(
            df_allvalues[["PATID", "inclusions", "inclusions_date"]]
        )
        df_drug_exclusion_exposure = pd.DataFrame(
            df_allvalues[["PATID", "exclusions", "exclusions_date"]]
        )
        df_drug_exposure_entry.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        df_drug_inclusion_exposure.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        df_drug_exclusion_exposure.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        df_drug_exposure = pd.concat(
            [
                df_drug_exposure_entry,
                df_drug_inclusion_exposure,
                df_drug_exclusion_exposure,
            ]
        )

        schema_drug_exposure = {
            "PATID": str,
            "PRODCODEID": str,
            "ISSUEDATE": datetime.date,
        }
        drug_exposure_table = DrugExposureTableForTests(
            self.con.create_table(
                "DRUG_EXPOSURE", df_drug_exposure, schema=schema_drug_exposure
            )
        )
        return {
            "PERSON": person_table,
            "DRUG_EXPOSURE": drug_exposure_table,
        }


class SimpleCohortWithInclusionAndExclusionSeparateTablesTestGenerator(
    CohortTestGenerator
):
    def define_cohort(self):
        entry = CodelistPhenotype(
            return_date="first",
            codelist=Codelist(["d1"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
        )

        i1 = CodelistPhenotype(
            name="prior_et_usage",
            codelist=Codelist(["i1"]).copy(use_code_type=False),
            domain="CONDITION_OCCURRENCE",
            relative_time_range=RelativeTimeRangeFilter(
                when="before", min_days=GreaterThanOrEqualTo(0)
            ),
        )

        e4 = CodelistPhenotype(
            name="prior_et_usage",
            codelist=Codelist(["e4"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
            relative_time_range=RelativeTimeRangeFilter(
                when="before", min_days=GreaterThanOrEqualTo(0)
            ),
        )

        return Cohort(
            name="test_simple_cohort_with_inclusions",
            entry_criterion=entry,
            inclusions=[i1],
            exclusions=[e4],
        )

    def generate_dummy_input_data(self):
        values = [
            {
                "name": "entry",
                "values": ["d1", "d4"],
            },
            {
                "name": "entry_date",
                "values": [datetime.date(2020, 1, 1), datetime.date(2010, 1, 1)],
            },
            {"name": "exclusions", "values": ["e1", "e4"]},
            {
                "name": "exclusions_date",
                "values": [datetime.date(2019, 4, 1)],
            },
            {"name": "inclusions", "values": ["i1", "i4"]},
            {
                "name": "inclusions_date",
                "values": [datetime.date(2019, 4, 1)],
            },
        ]
        return generate_dummy_cohort_data(values)

    def define_expected_output(self):
        df_expected_index = pd.DataFrame()
        df_expected_index["PERSON_ID"] = ["P0"]
        test_infos = {
            "index": df_expected_index,
        }
        return test_infos

    def define_mapped_tables(self):
        self.con = ibis.duckdb.connect()
        df_allvalues = self.generate_dummy_input_data()

        # create dummy person table
        df_person = pd.DataFrame(df_allvalues[["PATID"]])
        df_person["YOB"] = 1
        df_person["GENDER"] = 1
        df_person["ACCEPTABLE"] = 1
        schema_person = {"PATID": str, "YOB": int, "GENDER": int, "ACCEPTABLE": int}
        person_table = PersonTableForTests(
            self.con.create_table("PERSON", df_person, schema=schema_person)
        )
        # create drug exposure table
        df_drug_exposure_entry = pd.DataFrame(
            df_allvalues[["PATID", "entry", "entry_date"]]
        )
        df_drug_exclusion_exposure = pd.DataFrame(
            df_allvalues[["PATID", "exclusions", "exclusions_date"]]
        )
        df_drug_exposure_entry.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        df_drug_exclusion_exposure.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]

        df_drug_exposure = pd.concat(
            [df_drug_exposure_entry, df_drug_exclusion_exposure]
        )

        # create condition occurrence table
        df_inclusion = pd.DataFrame(
            df_allvalues[["PATID", "inclusions", "inclusions_date"]]
        )
        df_inclusion.columns = ["PATID", "MEDCODEID", "OBSDATE"]

        schema_drug_exposure = {
            "PATID": str,
            "PRODCODEID": str,
            "ISSUEDATE": datetime.date,
        }
        drug_exposure_table = DrugExposureTableForTests(
            self.con.create_table(
                "DRUG_EXPOSURE", df_drug_exposure, schema=schema_drug_exposure
            )
        )
        schema_condition_occurrence = {
            "PATID": str,
            "MEDCODEID": str,
            "OBSDATE": datetime.date,
        }
        condition_occurrence_table = ConditionOccurenceTableForTests(
            self.con.create_table(
                "CONDITION_OCCURRENCE", df_inclusion, schema=schema_condition_occurrence
            )
        )
        return {
            "PERSON": person_table,
            "DRUG_EXPOSURE": drug_exposure_table,
            "CONDITION_OCCURRENCE": condition_occurrence_table,
        }


class SimpleCohortWithContinuousCoverageInclusionTestGenerator(
    SimpleCohortWithExclusionTestGenerator
):
    def define_cohort(self):
        entry = CodelistPhenotype(
            return_date="first",
            codelist=Codelist(["d1"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
        )

        i1 = CodelistPhenotype(
            name="prior_et_usage",
            codelist=Codelist(["i1"]).copy(use_code_type=False),
            domain="CONDITION_OCCURRENCE",
            relative_time_range=RelativeTimeRangeFilter(
                when="before", min_days=GreaterThanOrEqualTo(0)
            ),
        )

        e4 = CodelistPhenotype(
            name="prior_et_usage",
            codelist=Codelist(["e4"]).copy(use_code_type=False),
            domain="DRUG_EXPOSURE",
            relative_time_range=RelativeTimeRangeFilter(
                when="before", min_days=GreaterThanOrEqualTo(0)
            ),
        )

        return Cohort(
            name="test_simple_cohort_with_inclusions",
            entry_criterion=entry,
            inclusions=[i1],
            exclusions=[e4],
        )

    def generate_dummy_input_data(self):
        values = [
            {
                "name": "entry",
                "values": ["d1", "d4"],
            },
            {
                "name": "entry_date",
                "values": [datetime.date(2020, 1, 1), datetime.date(2010, 1, 1)],
            },
            {"name": "exclusions", "values": ["e1", "e4"]},
            {
                "name": "exclusions_date",
                "values": [datetime.date(2019, 4, 1)],
            },
            {"name": "inclusions", "values": ["i1", "i4"]},
            {
                "name": "inclusions_date",
                "values": [datetime.date(2019, 4, 1)],
            },
        ]
        return generate_dummy_cohort_data(values)

    def define_expected_output(self):
        df_expected_index = pd.DataFrame()
        df_expected_index["PERSON_ID"] = ["P0"]
        test_infos = {
            "index": df_expected_index,
        }
        return test_infos

    def define_mapped_tables(self):
        self.con = ibis.duckdb.connect()
        df_allvalues = self.generate_dummy_input_data()

        # create dummy person table
        df_person = pd.DataFrame(df_allvalues[["PATID"]])
        df_person["YOB"] = 1
        df_person["GENDER"] = 1
        df_person["ACCEPTABLE"] = 1
        schema_person = {"PATID": str, "YOB": int, "GENDER": int, "ACCEPTABLE": int}
        person_table = PersonTableForTests(
            self.con.create_table("PERSON", df_person, schema=schema_person)
        )
        # create drug exposure table
        df_drug_exposure_entry = pd.DataFrame(
            df_allvalues[["PATID", "entry", "entry_date"]]
        )
        df_drug_exclusion_exposure = pd.DataFrame(
            df_allvalues[["PATID", "exclusions", "exclusions_date"]]
        )
        df_drug_exposure_entry.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        df_drug_exclusion_exposure.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]

        df_drug_exposure = pd.concat(
            [df_drug_exposure_entry, df_drug_exclusion_exposure]
        )

        # create condition occurrence table
        df_inclusion = pd.DataFrame(
            df_allvalues[["PATID", "inclusions", "inclusions_date"]]
        )
        df_inclusion.columns = ["PATID", "MEDCODEID", "OBSDATE"]

        schema_drug_exposure = {
            "PATID": str,
            "PRODCODEID": str,
            "ISSUEDATE": datetime.date,
        }
        drug_exposure_table = DrugExposureTableForTests(
            self.con.create_table(
                "DRUG_EXPOSURE", df_drug_exposure, schema=schema_drug_exposure
            )
        )
        schema_condition_occurrence = {
            "PATID": str,
            "MEDCODEID": str,
            "OBSDATE": datetime.date,
        }
        condition_occurrence_table = ConditionOccurenceTableForTests(
            self.con.create_table(
                "CONDITION_OCCURRENCE", df_inclusion, schema=schema_condition_occurrence
            )
        )
        return {
            "PERSON": person_table,
            "DRUG_EXPOSURE": drug_exposure_table,
            "CONDITION_OCCURRENCE": condition_occurrence_table,
        }


def test_simple_cohort_with_exclusion():
    g = SimpleCohortWithExclusionTestGenerator()
    g.run_tests()


def test_simple_cohort_with_exclusion_and_study_period():
    g = SimpleCohortWithExclusionAndStudyPeriodTestGenerator()
    g.run_tests()


def test_simple_cohort_with_exclusions_postindex():
    g = SimpleCohortWithExclusionPostIndexTestGenerator()
    g.run_tests()


def test_simple_cohort_with_inclusions_postindex():
    g = SimpleCohortWithInclusionTestGenerator()
    g.run_tests()


def test_simple_cohort_with_inclusions_and_exclusions():
    g = SimpleCohortWithInclusionAndExclusionTestGenerator()
    g.run_tests()


def test_simple_cohort_with_inclusions_and_exclusion_separate_table():
    g = SimpleCohortWithInclusionAndExclusionSeparateTablesTestGenerator()
    g.run_tests()


if __name__ == "__main__":
    # test_simple_cohort_with_exclusion()
    # test_simple_cohort_with_exclusion_and_study_period()
    # test_simple_cohort_with_exclusions_postindex()
    # test_simple_cohort_with_inclusions_postindex()
    # test_simple_cohort_with_inclusions_and_exclusions()
    test_simple_cohort_with_inclusions_and_exclusion_separate_table()



from typing import List
import pandas as pd
from .reporter import Reporter


class InExCounts(Reporter):
    """
    Get counts of inclusion and exclusion criteria

    """

    def execute(self, cohort: "Cohort") -> pd.DataFrame:
        self.cohort = cohort
        self.df_counts_inclusion = self.get_counts_for_phenotypes(
            self.cohort.inclusions, "inclusion"
        )
        self.df_counts_exclusion = self.get_counts_for_phenotypes(
            self.cohort.exclusions, "exclusion"
        )
        return pd.concat([self.df_counts_inclusion, self.df_counts_exclusion])

    def get_counts_for_phenotypes(
        self, phenotypes: List["Phenotype"], category: str = None
    ):
        ds = []
        for pt in phenotypes:
            d = {
                "phenotype": pt.name,
                "n": pt.table.select("PERSON_ID").distinct().count().to_pandas(),
            }
            ds.append(d)
        _df = pd.DataFrame.from_records(ds)
        if category is not None:
            _df["category"] = category
        return _df



from typing import Union, Optional, List
import math
import os
from lifelines import KaplanMeierFitter
from lifelines.plotting import add_at_risk_counts

import matplotlib.pyplot as plt
import ibis

from phenex.reporting import Reporter
from phenex.filters import ValueFilter
from phenex.util import create_logger

logger = create_logger(__name__)


class TimeToEvent(Reporter):
    """
    Perform a time to event analysis.

    The time_to_event table is first generated, after which, by default, a Kaplan Meier plot is generated. The time_to_event table contains one row per patient and then multiple columns containing

    ### Dates
    1. the index date for each patient
    2. the dates of all outcomes or NULL if they did not occur
    3. the dates of all right censoring events or NULL if they did not occur
    4. the date of the end of the study, if provided
    5. the days from index to the dates provided above
    6. the

    | Column | Description |
    | --- | --- |
    | `date` columns | The EVENT_DATE of every 1. cohort outcome phenotype and 2. the right censoring phenotypes, if provided. The column name for the respective EVENT_DATE is the name of the phenotype. 3. Additionally, the date of the end of study period is present, if provided. |
    | `days_to_event` columns | For each `date` column, the number of days from index_date to the `date` column. These columns begin with “DAYS_TO_” and the name of the `date` column. |
    | `date_first_event` columns | For each outcome phenotype, the `date_first_event` column is added titled “DATE_FIRST_EVENT_{name of outcome phenotype}”. This is the first date that occurs post index, whether that outcome, a right censoring event, or the end of study period. |
    | `days_to_first_event` columns | For each outcome phenotype, the `days_to_first_event` column is added titled “DAYS_FIRST_EVENT_{name of outcome phenotype}”. This is the days from index_date to the `date_first_event` column. |
    | `indicator` columns | For each outcome phenotype, the `indicator` column is added titled “INDICATOR_{name of outcome phenotype}”. This has a value of 1 if the first event was the outcome phenotype, or a 0 if the first event was a right censoring event or the end of study period. |

    Parameters:
        right_censor_phenotypes: A list of phenotypes that should be used as right censoring events. Suggested are death and end of followup.
        end_of_study_period: A datetime defining the end of study period.
    """

    def __init__(
        self,
        right_censor_phenotypes: Optional[List["Phenotype"]] = None,
        end_of_study_period: Optional["datetime"] = None,
    ):
        self.right_censor_phenotypes = right_censor_phenotypes
        self.end_of_study_period = end_of_study_period
        self._date_column_names = None

    def execute(self, cohort: "Cohort"):
        """
        Execute the time to event analysis for a provided cohort. This will generate a table with all necessary cohort outcome event dates and right censoring event dates. Following execution, a Kaplan Meier curve will be generated.

        Parameters:
            cohort: The cohort for which the time to event analysis should be performed.
        """
        self.cohort = cohort
        self._execute_right_censoring_phenotypes(self.cohort)

        table = cohort.index_table.mutate(
            INDEX_DATE=self.cohort.index_table.EVENT_DATE
        ).select(["PERSON_ID", "INDEX_DATE"])
        table = self._append_date_events(table)
        table = self._append_days_to_event(table)
        table = self._append_date_and_days_to_first_event(table)
        self.table = table
        logger.info("time to event finished execution")
        self.plot_multiple_kaplan_meier()

    def _execute_right_censoring_phenotypes(self, cohort):
        for phenotype in self.right_censor_phenotypes:
            phenotype.execute(cohort.subset_tables_index)

    def _append_date_events(self, table):
        """
        Append a column for all necessary event dates. This includes :
        1. the date of all outcome phenotypes; column name is name of phenotype
        2. the date of all right censor phenotypes; column name is name of phenotype
        3. date of end of study period; column name is END_OF_STUDY_PERIOD
        Additionally, this method populates _date_column_names with the name of all date columns appended here.
        """
        table = self._append_dates_for_phenotypes(table, self.cohort.outcomes)
        table = self._append_dates_for_phenotypes(table, self.right_censor_phenotypes)
        self._date_column_names = [
            x.name.upper() for x in self.cohort.outcomes + self.right_censor_phenotypes
        ]
        if self.end_of_study_period is not None:
            table = table.mutate(
                END_OF_STUDY_PERIOD=ibis.literal(self.end_of_study_period)
            )
            self._date_column_names.append("END_OF_STUDY_PERIOD")
        return table

    def _append_dates_for_phenotypes(self, table, phenotypes):
        """
        Generic method that adds the EVENT_DATE for a list of phenotypes

        For example, if three phenotypes are provided, named pt1, pt2, pt3, three new columns pt1, pt2, pt3 are added each populated with the EVENT_DATE of the respective phenotype.
        """
        for _phenotype in phenotypes:
            logger.info(f"appending dates for { _phenotype.name}")
            join_table = _phenotype.table.select(["PERSON_ID", "EVENT_DATE"]).distinct()
            # rename event_date to the right_censor_phenotype's name
            join_table = join_table.mutate(
                **{_phenotype.name.upper(): join_table.EVENT_DATE}
            )
            # select just person_id and event_date for current phenotype
            join_table = join_table.select(["PERSON_ID", _phenotype.name.upper()])
            # perform the join
            table = table.join(
                join_table, table.PERSON_ID == join_table.PERSON_ID, how="left"
            ).drop("PERSON_ID_right")
        return table

    def _append_days_to_event(self, table):
        """
        Calculates the days to each EVENT_DATE column found in _date_column_names. New columm names are "DAYS_TO_{date column name}".
        """
        for column_name in self._date_column_names:
            logger.info(f"appending time to event for {column_name}")
            DAYS_TO_EVENT = table[column_name].delta(table.INDEX_DATE, "day")
            table = table.mutate(**{f"DAYS_TO_{column_name}": DAYS_TO_EVENT})
        return table

    def _append_date_and_days_to_first_event(self, table):
        """
        For each outcome phenotype, determines which event occurred first, whether the outcome, a right censoring event, or the end of study period. Adds an indicator column whether the first event is the outcome.
        """
        for phenotype in self.cohort.outcomes:
            # Subset the columns from which the minimum date should be determined; this is the outcome of interest, all right censoring events, and end of study period.
            cols = [phenotype.name.upper()] + [
                x.name.upper() for x in self.right_censor_phenotypes
            ]

            # Create a proper minimum date calculation that handles nulls correctly
            # Start with a very large date as the initial minimum
            min_date_expr = ibis.literal(self.end_of_study_period)

            # For each column, update the minimum if the column has a valid (non-null) date that's smaller
            for col in cols:
                min_date_expr = (
                    ibis.case()
                    .when(
                        table[col].notnull() & (table[col] < min_date_expr), table[col]
                    )
                    .else_(min_date_expr)
                    .end()
                )

            min_date_column = min_date_expr

            # Adding the new column to the table
            table = table.mutate(min_date=min_date_column)

            # Adding the new column to the table
            column_name_date_first_event = f"DATE_FIRST_EVENT_{phenotype.name.upper()}"
            table = table.mutate(**{column_name_date_first_event: min_date_column})
            DAYS_FIRST_EVENT = table[column_name_date_first_event].delta(
                table.INDEX_DATE, "day"
            )
            table = table.mutate(
                **{f"DAYS_FIRST_EVENT_{phenotype.name.upper()}": DAYS_FIRST_EVENT}
            )
            # Adding an indicator for whether the first event was the outcome or a censoring event
            table = table.mutate(
                **{
                    f"INDICATOR_{phenotype.name.upper()}": ibis.ifelse(
                        table[phenotype.name.upper()]
                        == table[f"DATE_FIRST_EVENT_{phenotype.name.upper()}"],
                        1,
                        0,
                    )
                }
            )
        return table

    def plot_multiple_kaplan_meier(
        self,
        xlim: Optional[List[int]] = None,
        ylim: Optional[List[int]] = None,
        n_cols: int = 3,
        outcome_indices: Optional[List[int]] = None,
        path_dir: Optional[str] = None,
    ):
        """
        For each outcome, plot a kaplan meier curve.
        """
        # subset for current codelist
        phenotypes = self.cohort.outcomes
        if outcome_indices is not None:
            phenotypes = [
                x for i, x in enumerate(self.cohort.outcomes) if i in outcome_indices
            ]
        n_rows = math.ceil(len(phenotypes) / n_cols)
        fig, axes = plt.subplots(n_rows, n_cols, sharey=True, sharex=True)

        for i, phenotype in enumerate(phenotypes):
            kmf = self.fit_kaplan_meier_for_phenotype(phenotype)
            if n_rows > 1 and n_cols > 1:
                ax = axes[int(i / n_cols), i % n_cols]
            else:
                ax = axes[i]
            ax.set_title(phenotype.name)
            if xlim is not None:
                ax.set_xlim(xlim)
            if ylim is not None:
                ax.set_ylim(ylim)
            kmf.plot(ax=ax)
            ax.grid(color="gray", linestyle="-", linewidth=0.1)

        if path_dir is not None:
            path = os.path.join(path_dir, f"KaplanMeierPanelFor_{self.cohort.name}.svg")
            plt.savefig(path, dpi=150)
        plt.show()

    def plot_single_kaplan_meier(
        self,
        outcome_index: int = 0,
        xlim: Optional[List[int]] = None,
        ylim: Optional[List[int]] = None,
        path_dir: Optional[str] = None,
    ):
        """
        For each outcome, plot a kaplan meier curve.
        """
        # subset for current codelist
        phenotype = self.cohort.outcomes[outcome_index]
        fig, ax = plt.subplots(1, 1)

        kmf = self.fit_kaplan_meier_for_phenotype(phenotype)

        ax.set_title(f"Kaplan Meier for outcome : {phenotype.name}")
        kmf.plot(ax=ax)
        add_at_risk_counts(kmf, ax=ax)
        plt.tight_layout()
        ax.grid(color="gray", linestyle="-", linewidth=0.1)
        if xlim is not None:
            ax.set_xlim(xlim)
        if ylim is not None:
            ax.set_ylim(ylim)

        if path_dir is not None:
            path = os.path.join(
                path_dir, f"KaplanMeier_{self.cohort.name}_{phenotype.name}.svg"
            )
            plt.savefig(path, dpi=150)
        plt.show()

    def fit_kaplan_meier_for_phenotype(self, phenotype):
        indicator = f"INDICATOR_{phenotype.name.upper()}"
        durations = f"DAYS_FIRST_EVENT_{phenotype.name.upper()}"
        _sdf = self.table.select([indicator, durations])
        _df = _sdf.to_pandas()
        kmf = KaplanMeierFitter(label=phenotype.name)
        kmf.fit(durations=_df[durations], event_observed=_df[indicator])
        return kmf



from typing import Union, List, Dict, Optional
from phenex.phenotypes.phenotype import Phenotype
from phenex.filters.date_filter import ValueFilter
from phenex.tables import is_phenex_code_table, PHENOTYPE_TABLE_COLUMNS, PhenotypeTable
from phenex.phenotypes.functions import attach_anchor_and_get_reference_date
from ibis import _
from ibis.expr.types.relations import Table
import ibis


class TimeRangePhenotype(Phenotype):
    """
    As the name implies, TimeRangePhenotype is designed for working with time ranges. If the input data has a start and an end date, use TimeRangePhenotype to identify other events (or patients) that occur within this time range. The most common use case of this is working with 'health insurance coverage' data i.e. on 'OBSERVATION_PERIOD' table. These tables have one or many rows per patient with the start of coverage and end of coverage i.e. domains compatible with TimeRangePhenotype require a START_DATE and an END_DATE column. At it's simplest, TimeRangePhenotype identifies patients who have their INDEX_DATE (or other anchor date of interest) within this time range. Additionally, a minimum or maximum number of days from the anchor date to the beginning/end of the time range can be defined. The returned Phenotype has the following interpretation:

    DATE: If relative_time_range.when='before', then DATE is the beginning of the coverage period containing the anchor phenotype. If relative_time_range.when='after', then DATE is the end of the coverage period containing the anchor date.
    VALUE: Coverage (in days) relative to the anchor date. By convention, always non-negative.

    There are two primary use cases for TimeRangePhenotype:
        1. Identify patients with some minimum duration of coverage prior to anchor_phenotype date e.g. "identify patients with 1 year of continuous coverage prior to index date"
        2. Determine the date of loss to followup (right censoring) i.e. the duration of coverage after the anchor_phenotype event

    ## Data for TimeRangePhenotype
    This phenotype requires a table with PersonID and a coverage start date and end date. Depending on the datasource used, this information is a separate ObservationPeriod table or found in the PersonTable. Use an PhenexObservationPeriodTable to map required coverage start and end date columns. For tables with overlapping time ranges, use the CombineOverlappingPeriods derived table to combine time ranges into a single time range.

    | PersonID    |   startDate          |   endDate          |
    |-------------|----------------------|--------------------|
    | 1           |   2009-01-01         |   2010-01-01       |
    | 2           |   2008-01-01         |   2010-01-02       |

    One assumption that is made by TimeRangePhenotype is that there are **NO overlapping coverage periods**.

    Parameters:
        name: The name of the phenotype.
        domain: The domain of the phenotype. Default is 'observation_period'.
        relative_time_range: Filter returned persons based on the duration of coverage in days. The relative_time_range.anchor_phenotype defines the reference date with respect to calculate coverage. In typical applications, the anchor phenotype will be the entry criterion. The relative_time_range.when 'before', 'after'. If before, the return date is the start of the coverage period containing the anchor_phenotype. If after, the return date is the end of the coverage period containing the anchor_phenotype.
        allow_null_end_date: TimeRangePhenotype checks that anchor date is within the time range of interest. This requires that the start date is not null, and the end date is either null or after the anchor date. If you want to require that the end date is not null, set allow_null_end_date to False.

    Example:
    ```python
    # make sure to create an entry phenotype, for example 'atrial fibrillation diagnosis'
    entry_phenotype = CodelistPhenotype(...)
    # one year continuous coverage prior to index
    one_year_coverage = TimeRangePhenotype(
        relative_time_range = RelativeTimeRangeFilter(
            min_days=GreaterThanOrEqualTo(365),
            anchor_phenotype = entry_phenotype,
            when = 'before',
        ),
    )
    # determine the date of loss to followup
    loss_to_followup = TimeRangePhenotype(
        relative_time_range = RelativeTimeRangeFilter(
            anchor_phenotype = entry_phenotype
            when = 'after',
        )
    )

    # determine the date when a drug was discontinued
    drug_discontinuation = TimeRangePhenotype(
        relative_time_range = RelativeTimeRangeFilter(
            anchor_phenotype = entry_phenotype
            when = 'after',
        )
    )
    ```
    """

    def __init__(
        self,
        name: Optional[str] = "TIME_RANGE",
        domain: Optional[str] = "OBSERVATION_PERIOD",
        relative_time_range: Optional["RelativeTimeRangeFilter"] = None,
        allow_null_end_date: bool = True,
        **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self.domain = domain
        self.relative_time_range = relative_time_range
        self.allow_null_end_date = allow_null_end_date
        if self.relative_time_range is not None:
            if self.relative_time_range.anchor_phenotype is not None:
                self.children.append(self.relative_time_range.anchor_phenotype)

    def _execute(self, tables: Dict[str, Table]) -> PhenotypeTable:
        table = tables[self.domain]
        table, reference_column = attach_anchor_and_get_reference_date(
            table, self.relative_time_range.anchor_phenotype
        )

        # Ensure that the observation period includes anchor date
        # Allow END_DATE to be null (ongoing periods) if allow_null_end_date is True
        if self.allow_null_end_date:
            table = table.filter(
                (table.START_DATE <= reference_column)
                & ((reference_column <= table.END_DATE) | (table.END_DATE.isnull()))
            )
        else:
            table = table.filter(
                (table.START_DATE <= reference_column)
                & (reference_column <= table.END_DATE)
            )

        if (
            self.relative_time_range is None
            or self.relative_time_range.when == "before"
        ):
            VALUE = reference_column.delta(table.START_DATE, "day")
            EVENT_DATE = table.START_DATE
        else:
            VALUE = table.END_DATE.delta(reference_column, "day")
            EVENT_DATE = table.END_DATE

        table = table.mutate(VALUE=VALUE, EVENT_DATE=EVENT_DATE)

        if self.relative_time_range is not None:
            value_filter = ValueFilter(
                min_value=self.relative_time_range.min_days,
                max_value=self.relative_time_range.max_days,
                column_name="VALUE",
            )
            ibis.options.interactive = True
            table = value_filter.filter(table)

        return table



import pandas as pd

from .reporter import Reporter
from phenex.util import create_logger

logger = create_logger(__name__)


class Waterfall(Reporter):
    """
    A waterfall diagram, also known as an attrition table, shows how inclusion/exclusion criteria contribute to a final population size. Each inclusion/exclusion criteria is a row in the table, and the number of patients remaining after applying that criteria are shown on that row.

    | Column name | Description |
    | --- | --- |
    | Type | The type of the phenotype, either entry, inclusion or exclusion |
    | Name | The name of entry, inclusion or exclusion criteria |
    | N | The absolute number of patients that fulfill that phenotype. For the entry criterium this is the absolute number in the dataset. For inclusion/exclusion criteria this is the number of patients that fulfill the entry criterium AND the phenotype and that row. |
    | Remaining | The number of patients remaining in the cohort after sequentially applying the inclusion/exclusion criteria in the order that they are listed in this table. |
    | % | The percentage of patients who fulfill the entry criterion who are remaining in the cohort after application of the phenotype on that row |
    | Delta | The change in number of patients that occurs by applying the phenotype on that row. |

    """

    def execute(self, cohort: "Cohort") -> pd.DataFrame:
        self.cohort = cohort
        logger.debug(f"Beginning execution of waterfall. Calculating N patents")
        N = (
            cohort.index_table.filter(cohort.index_table.BOOLEAN == True)
            .select("PERSON_ID")
            .distinct()
            .count()
            .execute()
        )
        logger.debug(f"Cohort has {N} patients")
        self.ds = []

        table = cohort.entry_criterion.table
        N_entry = table.count().execute()
        self.ds.append(
            {
                "Type": "entry",
                "Name": (
                    cohort.entry_criterion.display_name
                    if self.pretty_display
                    else cohort.entry_criterion.name
                ),
                "N": N_entry,
                "Remaining": table.count().execute(),
            }
        )

        for inclusion in cohort.inclusions:
            table = self.append_phenotype_to_waterfall(table, inclusion, "inclusion")

        for exclusion in cohort.exclusions:
            table = self.append_phenotype_to_waterfall(table, exclusion, "exclusion")

        self.ds.append(
            {
                "Type": "final_cohort",
                "Name": "",
                "N": None,
                "Remaining": N,
            }
        )
        self.ds = self.append_delta(self.ds)

        # create dataframe with phenotype counts
        self.df = pd.DataFrame(self.ds)

        # calculate percentage of entry criterion
        self.df["%"] = self.df["Remaining"] / N_entry * 100
        self.df = self.df.round(self.decimal_places)

        if self.pretty_display:
            self.create_pretty_display()

        # Do final column selection
        self.df = self.df[["Type", "Name", "N", "Remaining", "%", "Delta"]]

        return self.df

    def append_phenotype_to_waterfall(self, table, phenotype, type):
        if type == "inclusion":
            table = table.inner_join(
                phenotype.table, table["PERSON_ID"] == phenotype.table["PERSON_ID"]
            )
        elif type == "exclusion":
            table = table.filter(~table["PERSON_ID"].isin(phenotype.table["PERSON_ID"]))
        else:
            raise ValueError("type must be either inclusion or exclusion")
        logger.debug(f"Starting {type} criteria {phenotype.name}")
        self.ds.append(
            {
                "Type": type,
                "Name": (
                    phenotype.display_name if self.pretty_display else phenotype.name
                ),
                "N": phenotype.table.select("PERSON_ID").distinct().count().execute(),
                "Remaining": table.select("PERSON_ID").distinct().count().execute(),
            }
        )
        logger.debug(
            f"Finished {type} criteria {phenotype.name}: N = {self.ds[-1]['N']} waterfall = {self.ds[-1]['Remaining']}"
        )
        return table.select("PERSON_ID")

    def append_delta(self, ds):
        ds[0]["Delta"] = None
        for i in range(1, len(ds) - 1):
            d_current = ds[i]
            d_previous = ds[i - 1]
            d_current["Delta"] = d_current["Remaining"] - d_previous["Remaining"]
        return ds

    def create_pretty_display(self):
        # cast counts to integer and to str, so that we can display without 'NaNs'
        self.df["N"] = self.df["N"].astype("Int64").astype(str)
        self.df["Delta"] = self.df["Delta"].astype("Int64").astype(str)

        # Replace NAs and None values with empty strings for display
        self.df = self.df.replace("<NA>", "")

        # create a sparse 'type' column name where inclusion/exclusion only appear once (instead of repeated on each row)
        previous_type = None
        sparse_types = []
        for _type in self.df["Type"].values:
            if _type != previous_type:
                sparse_types.append(_type)
                previous_type = _type
            else:
                sparse_types.append("")
        self.df["Type"] = sparse_types



from typing import Optional
from phenex.phenotypes import MeasurementPhenotype, Phenotype
from phenex.filters.value import Value, GreaterThanOrEqualTo
from phenex.filters.value_filter import ValueFilter
from phenex.tables import PHENOTYPE_TABLE_COLUMNS, PhenotypeTable
from phenex.aggregators.aggregator import First, Last, ValueAggregator, DailyMedian

from ibis import _


class MeasurementChangePhenotype(Phenotype):
    """
    MeasurementChangePhenotype looks for changes in the value of a MeasurementPhenotype within a certain time period. Returns EVENT_DATE as either the date of the first or second MeasurementPhenotype event and VALUE as the observed change in the underlying MeasurementPhenotype's VALUE.

    Parameters:
        name: The name of the phenotype.
        phenotype: The measurement phenotype to look for changes.
        min_change: The minimum change in the measurement value to look for.
        direction: Whether the min_change / max_change are to be interpreted as an 'increase' in the value or a 'decrease' in the value. For example, min_change = GreatThan(0), max_change = LessThan(3), direction = 'decrease' looks for a decrease of the value between 0 and 3.
        min_days_apart: The minimum number of days between the measurements. FIXME Currently unable to detect measurement changes on the same day due to lack of intraday time resolution!
        max_days_apart: The maximum number of days between the measurements.
        component_date_select: Specifies which MeasurementPhenotype event to use for defining the date of the MeasurementChange ('first' or 'second')
        return_date: Specifies which MeasurementChangePhenotyp event to return among all events passing the filters ('first', 'last' or 'all')
        return_value: A ValueAggregator operation describing which value to return when more than one value per person pass all the filters. FIXME: return_value and return_date don't really place nicely together in this Phenotype. If you need this feature or experience breakages, reach out for support. This is a known issue but it was decided to put off fixing it.

    Example: Drop in Hemoglobin of at Least 2g/dL
        ```python
        hemoglobin = MeasurementPhenotype(
            name='hemoglobin_drop',
            codelist=hb_codes,
            domain='observation',
            relative_time_range=ONEYEAR_PREINDEX,
        )

        hemoglobin_drop = MeasurementChangePhenotype(
            phenotype=hemoglobin,
            min_change=GreaterThanOrEqualTo(2),
            max_days_apart=LessThanOrEqualTo(2),
            direction='decrease',
            component_date_select='second',
            return_date='first'
        )
        ```
    """

    def __init__(
        self,
        phenotype: MeasurementPhenotype,
        min_change: Value = None,
        max_change: Value = None,
        direction="increase",
        min_days_between: Value = GreaterThanOrEqualTo(0),
        max_days_between: Value = None,
        component_date_select="second",
        return_date="first",
        return_value: Optional[ValueAggregator] = None,
        **kwargs,
    ):
        super(MeasurementChangePhenotype, self).__init__(**kwargs)
        self.phenotype = phenotype
        self.min_change = min_change
        self.max_change = max_change
        self.direction = direction
        self.min_days_between = min_days_between
        self.max_days_between = max_days_between
        self.component_date_select = component_date_select
        self.return_date = return_date
        self.return_value = return_value
        self.children = [phenotype]
        if self.direction not in ["increase", "decrease"]:
            raise ValueError(
                f"Invalid choice for direction: {direction}. Must be 'increase' or 'decrease'."
            )
        if component_date_select not in ["first", "second"]:
            raise ValueError(
                f'component_date_select = {component_date_select} not supported, must be either "first" or "second"'
            )

    def _execute(self, tables) -> PhenotypeTable:
        # Execute the child phenotype to get the initial filtered table
        phenotype_table_1 = self.phenotype.table
        phenotype_table_2 = self.phenotype.table.view()
        # Create a self-join to compare each measurement with every other measurement
        joined_table = phenotype_table_1.join(
            phenotype_table_2,
            [
                phenotype_table_1.PERSON_ID == phenotype_table_2.PERSON_ID,
                (phenotype_table_1.EVENT_DATE != phenotype_table_2.EVENT_DATE)
                | (phenotype_table_1.VALUE != phenotype_table_2.VALUE),
            ],
            lname="{name}_1",
            rname="{name}_2",
        ).filter(_.EVENT_DATE_1 < _.EVENT_DATE_2)
        # Calculate the change in value and the days apart
        days_between = joined_table.EVENT_DATE_2.delta(joined_table.EVENT_DATE_1, "day")
        value_change = joined_table.VALUE_2 - joined_table.VALUE_1
        joined_table = joined_table.mutate(
            VALUE_CHANGE=value_change, DAYS_BETWEEN=days_between
        )

        # Filter to keep only those with at least min_change and within max_days_apart
        if self.direction == "decrease":
            if self.min_change:
                max_change = Value(
                    operator=self.min_change.operator.replace(">", "<"),
                    value=-self.min_change.value,
                )
            else:
                max_change = None
            if self.max_change:
                min_change = Value(
                    operator=self.max_change.operator.replace("<", ">"),
                    value=-self.max_change.value,
                )
            else:
                min_change = None
        else:
            min_change = self.min_change
            max_change = self.max_change

        value_filter = ValueFilter(
            min_value=min_change, max_value=max_change, column_name="VALUE_CHANGE"
        )
        filtered_table = value_filter.filter(joined_table)
        time_filter = ValueFilter(
            min_value=self.min_days_between,
            max_value=self.max_days_between,
            column_name="DAYS_BETWEEN",
        )
        filtered_table = time_filter.filter(filtered_table)

        # Determine the return date based on the return_date attribute
        if self.component_date_select == "first":
            filtered_table = filtered_table.mutate(
                EVENT_DATE=filtered_table.EVENT_DATE_1,
            )
        elif self.component_date_select == "second":
            filtered_table = filtered_table.mutate(
                EVENT_DATE=filtered_table.EVENT_DATE_2,
            )

        # Select the required columns
        filtered_table = filtered_table.mutate(
            PERSON_ID="PERSON_ID_1", VALUE="VALUE_CHANGE", BOOLEAN=True
        )

        # Handle the return_date attribute for each PERSON_ID using window functions
        if self.return_date == "first":
            filtered_table = First(reduce=False).aggregate(filtered_table)
        elif self.return_date == "last":
            filtered_table = Last(reduce=False).aggregate(filtered_table)

        if self.return_value is not None:
            filtered_table = self.return_value.aggregate(filtered_table)

        filtered_table = (
            filtered_table.mutate(BOOLEAN=True)
            .select(PHENOTYPE_TABLE_COLUMNS)
            .distinct()
        )

        return filtered_table



import os
import datetime

import pandas as pd
import ibis
import ibis.expr.datatypes as dt
from pyarrow import int8

from .util.check_equality import check_equality

import logging
from phenex.tables import *


class PhenotypeTestGenerator:
    """
    This class is a base class for all TestGenerators.

    FIXME Document how to subclass and use.
    """

    name_space = ""
    date_format = "%m-%d-%Y"
    test_values = False
    value_datatype = float
    test_date = False
    join_on = ["PERSON_ID"]

    def run_tests(self, verbose=False):
        self.verbose = verbose
        self.con = ibis.duckdb.connect()
        self._create_artifact_directory(self.name_space)
        self._create_input_data()
        self._run_tests()

    def define_input_tables(self):
        raise NotImplementedError

    def define_phenotype_tests(self):
        raise NotImplementedError

    def _create_artifact_directory(self, name_phenotype):
        path_artificats = os.path.join(
            os.path.dirname(__file__), "phenotypes/artifacts", name_phenotype
        )
        if not os.path.exists(path_artificats):
            os.makedirs(path_artificats)
        self.dirpaths = {
            "seeds": path_artificats,
            "input": os.path.join(path_artificats, "input_generated"),
            "expected": os.path.join(path_artificats, "expected"),
            "result": os.path.join(path_artificats, "result"),
        }
        for _path in self.dirpaths.values():
            if not os.path.exists(_path):
                os.makedirs(_path)

    def name_file(self, table_name):
        return f"{self.name_space}_{table_name}"

    def name_output_file(self, test_info):
        return f"{test_info['phenotype'].name}_output"

    def _create_input_data(self):
        self.domains = {}
        input_infos = self.define_input_tables()
        for input_info in input_infos:
            filename = self.name_file(input_info["name"]) + ".csv"
            path = os.path.join(self.dirpaths["input"], filename)
            # NOTE Depending on which database you are using, dates in the seed
            # data may have to be in a different format. It's best to keep your
            # seed data dates in pandas datetime format and let the
            # TestGenerator format them into strings.
            input_info["df"].to_csv(path, index=False, date_format=self.date_format)

            schema = {}
            for col in input_info["df"].columns:
                if "date" in col.lower():
                    schema[col] = datetime.date
                elif "value" in col.lower():
                    schema[col] = float
                else:
                    schema[col] = str

            table = self.con.create_table(
                input_info["name"], input_info["df"], schema=schema
            )
            if "type" in input_info.keys():
                table_type = input_info["type"]
            else:
                table_type = PhenexTable
                if input_info["name"].lower() in [
                    "condition_occurrence",
                ]:
                    table_type = CodeTable
                if input_info["name"].lower() in ["measurement"]:
                    table_type = MeasurementTable
                elif input_info["name"].lower() in ["person"]:
                    table_type = PhenexPersonTable

            self.domains[input_info["name"]] = table_type(table)

    def _run_tests(self):
        def df_from_test_info(test_info):
            df = pd.DataFrame()
            df["PERSON_ID"] = test_info["persons"]

            columnname_boolean = "boolean"
            columnname_date = "EVENT_DATE"
            columnname_value = "VALUE"

            df[columnname_boolean] = True

            if test_info.get("dates") is not None:
                df[columnname_date] = test_info["dates"]
            else:
                df[columnname_date] = None

            if test_info.get("values") is not None:
                df[columnname_value] = test_info["values"]
            else:
                df[columnname_value] = None

            return df

        self.test_infos = self.define_phenotype_tests()

        for test_info in self.test_infos:
            df = df_from_test_info(test_info)
            filename = self.name_output_file(test_info) + ".csv"
            path = os.path.join(self.dirpaths["expected"], filename)
            # NOTE Depending on which database you are using, dates in the seed
            # data may have to be in a different format. It's best to keep your
            # seed data dates in pandas datetime format and let the
            # TestGenerator format them into strings.
            df.to_csv(path, index=False, date_format=self.date_format)

            result_table = test_info["phenotype"].execute(self.domains)

            if self.verbose:
                ibis.options.interactive = True
                print("PRINTING THE SQL")
                ibis.to_sql(result_table)
                print(
                    f"Running test: {test_info['name']}\nExpected output:\n{df}\nActualOutput:\n{result_table}\n\n"
                )
            path = os.path.join(self.dirpaths["result"], filename)
            result_table.to_pandas().to_csv(
                path, index=False, date_format=self.date_format
            )

            schema = {}
            for col in df.columns:
                if "date" in col.lower():
                    schema[col] = datetime.date
                elif "value" in col.lower():
                    schema[col] = self.value_datatype
                elif "boolean" in col.lower():
                    schema[col] = bool
                else:
                    schema[col] = str

            expected_output_table = self.con.create_table(
                self.name_output_file(test_info), df, schema=schema
            )

            join_on = ["PERSON_ID"]
            if self.test_values:
                join_on.append("VALUE")
            if self.test_date:
                join_on.append("EVENT_DATE")
            check_equality(
                result_table,
                expected_output_table,
                test_name=test_info["name"],
                test_values=self.test_values,
                test_date=self.test_date,
                join_on=join_on,
            )



from typing import Dict, Union, Optional
from ibis.expr.types.relations import Table
from deepdiff import DeepDiff
from phenex.tables import (
    PhenotypeTable,
    PHENOTYPE_TABLE_COLUMNS,
    is_phenex_phenotype_table,
)
from phenex.util import create_logger
from phenex.util.serialization.to_dict import to_dict

logger = create_logger(__name__)


class Phenotype:
    """
    A phenotype is a description of the state of a person at a specific time.

    In Phenex, phenotypes are implemented using the Phenotype class. The Phenotype class is designed so that there is clear separation between the "what" from the "how". The "what" is expressed in the Phenotype init function: what codelists to use, what time range to include, constraints relative to other Phenotype's, visit detail information to include, etc. The "what" is meant to mirror how we normally talk about real-world data studies.

    The translation of this description in actual executable code (the "how") is handled via the `Phenotype.execute()` method. The execute method returns a PhenotypeTable - the realization of the defined Phenotype in a particular database. See `execute()` for details.

    All Phenotype's in Phenex derive from the Phenotype class.

    To subclass:
        1. Define the parameters required to compute the Phenotype in the `__init__()` interface.
        2. Within `__init__()`, define `self.children` - a list of Phenotype's which must be executed before the current Phenotype, allowing Phenotype's to be chained and executed recursively.
        3. Define `self._execute()`. The `self._execute()` method is reponsible for interpreting the input parameters to the Phenotype and returning the appropriate PhenotypeTable.
        4. Define tests in `phenex.test.phenotypes`! We demand a high level of test coverage for our code. High test coverage gives us confidence that our answers are correct and makes it easier to make changes to the code later on.

    Parameters:
        description: A plain text description of the phenotype.
    """

    def __init__(self, name: Optional[str] = None, description: Optional[str] = None):
        self.table = (
            None  # self.table is populated ONLY AFTER self.execute() is called!
        )
        self._name = name
        self.children = []  # List[Phenotype]
        self.description = description
        self._check_for_children()

    @property
    def name(self):
        if self._name is not None:
            return self._name.upper()
        return "PHENOTYPE"  # TODO replace with phenotype id when phenotype id is implemented

    @name.setter
    def name(self, name):
        self._name = name

    def execute(self, tables: Dict[str, Table]) -> PhenotypeTable:
        """
        Executes the phenotype computation for the current object and its children. This method recursively iterates over the children of the current object and calls their execute method if their table attribute is None.

        Args:
            tables (Dict[str, PhenexTable]): A dictionary mapping table names to PhenexTable objects. See phenex.mappers.DomainsDictionary.get_mapped_tables().

        Returns:
            table (PhenotypeTable): The resulting phenotype table containing the required columns. The PhenotypeTable will contain the columns: PERSON_ID, EVENT_DATE, VALUE. DATE is determined by the return_date parameter. VALUE is different for each phenotype. For example, AgePhenotype will return the age in the VALUE column. A MeasurementPhenotype will return the observed value for the measurement. See the specific phenotype of interest to understand more.
        """
        logger.info(f"Phenotype '{self.name}': executing...")
        for child in self.children:
            if child.table is None:
                logger.debug(
                    f"Phenotype {self.name}: executing child phenotype '{child.name}'..."
                )
                child.execute(tables)
            else:
                logger.debug(
                    f"Phenotype {self.name}: skipping already computed child phenotype '{child.name}'."
                )

        table = self._execute(tables).mutate(BOOLEAN=True)

        if not set(PHENOTYPE_TABLE_COLUMNS) <= set(table.columns):
            raise ValueError(
                f"Phenotype {self.name} must return columns {PHENOTYPE_TABLE_COLUMNS}. Found {table.columns}."
            )

        self.table = table.select(PHENOTYPE_TABLE_COLUMNS)
        # for some reason, having NULL datatype screws up writing the table to disk; here we make explicit cast
        if type(self.table.schema()["VALUE"]) == ibis.expr.datatypes.core.Null:
            self.table = self.table.cast({"VALUE": "float64"})

        assert is_phenex_phenotype_table(self.table)
        logger.info(f"Phenotype '{self.name}': execution completed.")
        return self.table

    @property
    def namespaced_table(self) -> Table:
        """
        A PhenotypeTable has generic column names 'person_id', 'boolean', 'event_date', and 'value'. The namespaced_table prepends the phenotype name to all of these columns. This is useful when joining multiple phenotype tables together.

        Returns:
            table (Table): The namespaced table for the current phenotype.
        """
        if self.table is None:
            raise ValueError("Phenotype has not been executed yet.")
        # since phenotypes may be executed multiple times (in an interactive setting for example), we must always get the namespaced table freshly from self.table
        new_column_names = {
            "PERSON_ID": "PERSON_ID",
            f"{self.name}_BOOLEAN": "BOOLEAN",
            f"{self.name}_EVENT_DATE": "EVENT_DATE",
            f"{self.name}_VALUE": "VALUE",
        }
        return self.table.rename(new_column_names)

    def _execute(self, tables: Dict[str, Table]):
        """
        Executes the phenotype processing logic.

        Args:
            tables (Dict[str, Table]): A dictionary where the keys are table names and the values are Table objects.

        Raises:
            NotImplementedError: This method should be implemented by subclasses.
        """
        raise NotImplementedError()

    def _check_for_children(self):
        for phenotype in self.children:
            if not isinstance(phenotype, Phenotype):
                raise ValueError("Dependent children must be of type Phenotype!")

    def __add__(
        self, other: Union["Phenotype", "ComputationGraph"]
    ) -> "ComputationGraph":
        return ComputationGraph(self, other, "+")

    def __radd__(
        self, other: Union["Phenotype", "ComputationGraph"]
    ) -> "ComputationGraph":
        return ComputationGraph(self, other, "+")

    def __sub__(
        self, other: Union["Phenotype", "ComputationGraph"]
    ) -> "ComputationGraph":
        return ComputationGraph(self, other, "-")

    def __mul__(
        self, other: Union["Phenotype", "ComputationGraph"]
    ) -> "ComputationGraph":
        return ComputationGraph(self, other, "*")

    def __rmul__(
        self, other: Union["Phenotype", "ComputationGraph"]
    ) -> "ComputationGraph":
        return ComputationGraph(self, other, "*")

    def __truediv__(
        self, other: Union["Phenotype", "ComputationGraph"]
    ) -> "ComputationGraph":
        return ComputationGraph(self, other, "/")

    def __and__(
        self, other: Union["Phenotype", "ComputationGraph"]
    ) -> "ComputationGraph":
        return ComputationGraph(self, other, "&")

    def __or__(
        self, other: Union["Phenotype", "ComputationGraph"]
    ) -> "ComputationGraph":
        return ComputationGraph(self, other, "|")

    def __invert__(self) -> "ComputationGraph":
        return ComputationGraph(self, None, "~")

    def __eq__(self, other) -> bool:
        diff = DeepDiff(self.to_dict(), other.to_dict(), ignore_order=True)
        if diff:
            logger.info(f"{self.__class__.__name__}s NOT equal")
            logger.info(diff)
            return False
        else:
            logger.debug(f"{self.__class__.__name__}s are equal")
            return True

    def get_codelists(self, to_pandas=False):
        codelists = []
        for child in self.children:
            codelists.extend(child.get_codelists())

        if to_pandas:
            import pandas as pd

            return pd.concat([x.to_pandas() for x in codelists]).drop_duplicates()
        return codelists

    def to_dict(self):
        return to_dict(self)

    @property
    def display_name(self):
        return self.name.replace("_", " ").lower().capitalize()


from typing import Dict, Union
from datetime import date
import ibis
from ibis.expr.types.relations import Table
from phenex.tables import PhenotypeTable, PHENOTYPE_TABLE_COLUMNS


class ComputationGraph:
    """
    ComputationGraph tracks arithmetic operations to be performed on two Phenotype objects.
    The actual execution of these operations is context-dependent and is handled by the
    responsible Phenotype class (ArithmeticPhenotype, ScorePhenotype, LogicPhenotype, etc.).
    """

    def __init__(
        self,
        left: Union["Phenotype", "ComputationGraph"],
        right: Union["Phenotype", "ComputationGraph", int, float, None],
        operator: str,
    ):
        self.table = None
        self.left = left
        self.right = right
        self.operator = operator
        self.children = (
            [left]
            if right is None or isinstance(right, (int, float))
            else [left, right]
        )

    def __add__(
        self, other: Union["Phenotype", "ComputationGraph"]
    ) -> "ComputationGraph":
        return ComputationGraph(self, other, "+")

    def __radd__(
        self, other: Union["Phenotype", "ComputationGraph"]
    ) -> "ComputationGraph":
        return ComputationGraph(self, other, "+")

    def __sub__(
        self, other: Union["Phenotype", "ComputationGraph"]
    ) -> "ComputationGraph":
        return ComputationGraph(self, other, "-")

    def __mul__(
        self, other: Union["Phenotype", "ComputationGraph"]
    ) -> "ComputationGraph":
        return ComputationGraph(self, other, "*")

    def __rmul__(
        self, other: Union["Phenotype", "ComputationGraph"]
    ) -> "ComputationGraph":
        return ComputationGraph(self, other, "*")

    def __truediv__(
        self, other: Union["Phenotype", "ComputationGraph"]
    ) -> "ComputationGraph":
        return ComputationGraph(self, other, "/")

    def __and__(
        self, other: Union["Phenotype", "ComputationGraph"]
    ) -> "ComputationGraph":
        return ComputationGraph(self, other, "&")

    def __or__(
        self, other: Union["Phenotype", "ComputationGraph"]
    ) -> "ComputationGraph":
        return ComputationGraph(self, other, "|")

    def __invert__(self) -> "ComputationGraph":
        return ComputationGraph(self, None, "~")

    def get_leaf_phenotypes(self):
        """
        A recursive function to extract all the leaf phenotypes from a computation graph.
        """

        def manage_node(node):
            if isinstance(node, ComputationGraph):
                return node.get_leaf_phenotypes()
            elif isinstance(node, Phenotype):
                return [node]
            return []

        phenotypes = []
        phenotypes.extend(manage_node(self.left))
        phenotypes.extend(manage_node(self.right))
        return phenotypes

    def get_value_expression(self, table, operate_on="boolean"):
        """
        A recursive function to build the full expression defined by a computation graph. A computation graph is a tree like structure with parents and children. The children can be either Phenotype objects, other ComputationGraph objects, or numerical values (int/float). The parents are the arithmetic operators that define the relationship between the children. This function recursively builds the expression by calling itself on the children and then applying the operator to the results.

        Args:
            table (Table): The table on which the value_expression is to be executed on. This must be the joined table that contains all the phenotypes contained within the computation graph.
            operate_on (str): Either 'boolean' or 'value', depending on whether the expression is to be evaluated using the phenotype boolean columns or value columns. See the comparison of composite phenotypes for more information.
        """

        def manage_node(node):
            if isinstance(node, ComputationGraph):
                return node.get_value_expression(table, operate_on)
            elif isinstance(node, Phenotype):
                if operate_on == "boolean":
                    return table[f"{node.name}_BOOLEAN"]
                return table[f"{node.name}_VALUE"]
            return node

        left = manage_node(self.left)
        right = manage_node(self.right)
        if self.operator == "+":
            return left + right
        elif self.operator == "-":
            return left - right
        elif self.operator == "*":
            return left * right
        elif self.operator == "/":
            return left / right
        else:
            raise ValueError(f"Operator {self.operator} not supported.")

    def get_boolean_expression(self, table, operate_on="boolean"):
        def manage_node(node):
            if isinstance(node, ComputationGraph):
                return node.get_boolean_expression(table, operate_on)
            elif isinstance(node, Phenotype):
                if operate_on == "boolean":
                    return table[f"{node.name}_BOOLEAN"]
                return table[f"{node.name}_VALUE"]
            return node

        left = manage_node(self.left)
        right = manage_node(self.right)

        if self.operator == "|":
            return left | right
        elif self.operator == "&":
            return left & right
        elif self.operator == "~":
            return ~(left)
        else:
            raise ValueError(f"Operator {self.operator} not supported.")

    def get_str(self):
        def manage_node(node):
            if isinstance(node, ComputationGraph):
                return node.get_str()
            elif isinstance(node, Phenotype):
                return node.name
            return str(node)

        left = manage_node(self.left)
        right = manage_node(self.right)
        return f"({left} {self.operator} {right})"

    def __str__(self):
        return self.get_str()

    def to_dict(self):
        return to_dict(self)
